from __future__ import annotations

import asyncio
import binascii
import functools
import hashlib
import inspect
import re
import time
from typing import Any, Callable, Dict, List, Literal, Match, Optional, Set, Tuple, Union, cast, overload

import aioamqp
import aioamqp.properties

from tomodachi import get_contextvar, logging
from tomodachi._exception import limit_exception_traceback
from tomodachi.helpers.execution_context import (
    decrease_execution_context_value,
    increase_execution_context_value,
    set_execution_context,
)
from tomodachi.helpers.middleware import execute_middlewares
from tomodachi.invoker import Invoker
from tomodachi.options import Options

MESSAGE_ENVELOPE_DEFAULT = "2594418c-5771-454a-a7f9-8f83ae82812a"
MESSAGE_PROTOCOL_DEFAULT = MESSAGE_ENVELOPE_DEFAULT  # deprecated
MESSAGE_ROUTING_KEY_PREFIX = "38f58822-25f6-458a-985c-52701d40dbbc"


class AmqpException(Exception):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self._log_level = kwargs.get("log_level") if kwargs and kwargs.get("log_level") else "INFO"


class AmqpInternalServiceError(AmqpException):
    pass


class AmqpInternalServiceErrorException(AmqpInternalServiceError):
    pass


class AmqpInternalServiceException(AmqpInternalServiceError):
    pass


class AmqpExclusiveQueueLockedException(AmqpException):
    pass


class AmqpTooManyConsumersException(AmqpException):
    pass


class AmqpConnectionException(AmqpException):
    pass


class AmqpChannelClosed(AmqpException):
    pass


class AmqpTransport(Invoker):
    channel: Any = None
    protocol: Any = None
    transport: Any = None
    exchange_name: str

    @overload
    @classmethod
    async def publish(
        cls,
        service: Any,
        data: Any,
        routing_key: str = "",
        exchange_name: str = "",
        wait: Literal[True] = True,
        *,
        message_envelope: Any = MESSAGE_ENVELOPE_DEFAULT,
        message_protocol: Any = MESSAGE_ENVELOPE_DEFAULT,  # deprecated
        routing_key_prefix: Optional[str] = MESSAGE_ROUTING_KEY_PREFIX,
        **kwargs: Any,
    ) -> None: ...

    @overload
    @classmethod
    async def publish(
        cls,
        service: Any,
        data: Any,
        routing_key: str,
        exchange_name: str,
        wait: Literal[False],
        *,
        message_envelope: Any = MESSAGE_ENVELOPE_DEFAULT,
        message_protocol: Any = MESSAGE_ENVELOPE_DEFAULT,  # deprecated
        routing_key_prefix: Optional[str] = MESSAGE_ROUTING_KEY_PREFIX,
        **kwargs: Any,
    ) -> asyncio.Task[None]: ...

    @overload
    @classmethod
    async def publish(
        cls,
        service: Any,
        data: Any,
        routing_key: str = "",
        exchange_name: str = "",
        *,
        wait: Literal[False],
        message_envelope: Any = MESSAGE_ENVELOPE_DEFAULT,
        message_protocol: Any = MESSAGE_ENVELOPE_DEFAULT,  # deprecated
        routing_key_prefix: Optional[str] = MESSAGE_ROUTING_KEY_PREFIX,
        **kwargs: Any,
    ) -> asyncio.Task[None]: ...

    @classmethod
    async def publish(
        cls,
        service: Any,
        data: Any,
        routing_key: str = "",
        exchange_name: str = "",
        wait: bool = True,
        *,
        message_envelope: Any = MESSAGE_ENVELOPE_DEFAULT,
        message_protocol: Any = MESSAGE_ENVELOPE_DEFAULT,  # deprecated
        routing_key_prefix: Optional[str] = MESSAGE_ROUTING_KEY_PREFIX,
        **kwargs: Any,
    ) -> Optional[asyncio.Task[None]]:
        if not cls.channel:
            await cls.connect(service, service.context)
        exchange_name = exchange_name or cls.exchange_name or "amq.topic"

        if message_envelope == MESSAGE_ENVELOPE_DEFAULT and message_protocol != MESSAGE_ENVELOPE_DEFAULT:
            # Fallback if deprecated message_protocol keyword is used
            message_envelope = message_protocol

        message_envelope = (
            getattr(service, "message_envelope", getattr(service, "message_protocol", None))
            if message_envelope == MESSAGE_ENVELOPE_DEFAULT
            else message_envelope
        )

        payload = data
        if message_envelope:
            build_message_func = getattr(message_envelope, "build_message", None)
            if build_message_func:
                payload = await asyncio.create_task(build_message_func(service, routing_key, data, **kwargs))

        async def _publish_message() -> None:
            properties: Dict = {}
            await cls._publish_message(
                routing_key, exchange_name, payload, properties, routing_key_prefix, service, service.context
            )

        if wait:
            return await asyncio.create_task(_publish_message())
        else:
            return asyncio.create_task(_publish_message())

    @classmethod
    async def _publish_message(
        cls,
        /,
        routing_key: str,
        exchange_name: str,
        payload: Any,
        properties: Dict,
        routing_key_prefix: Optional[str],
        service: Any,
        context: Dict,
    ) -> None:
        success = False
        while not success:
            try:
                await cls.channel.basic_publish(
                    str.encode(payload),
                    exchange_name,
                    cls.encode_routing_key(cls.get_routing_key(routing_key, context, routing_key_prefix)),
                    properties,
                )
                success = True
            except AssertionError:
                await cls.connect(service, context)

    @classmethod
    def get_routing_key(
        cls, routing_key: str, context: Dict, routing_key_prefix: Optional[str] = MESSAGE_ROUTING_KEY_PREFIX
    ) -> str:
        routing_key_prefix = (
            cls.options(context).amqp.routing_key_prefix
            if routing_key_prefix == MESSAGE_ROUTING_KEY_PREFIX
            else (routing_key_prefix or "")
        )
        if routing_key_prefix:
            return "{}{}".format(routing_key_prefix, routing_key)
        return routing_key

    @classmethod
    def get_routing_key_without_prefix(
        cls, routing_key: str, context: Dict, routing_key_prefix: Optional[str] = MESSAGE_ROUTING_KEY_PREFIX
    ) -> str:
        routing_key_prefix = (
            cls.options(context).amqp.routing_key_prefix
            if routing_key_prefix == MESSAGE_ROUTING_KEY_PREFIX
            else (routing_key_prefix or "")
        )
        if routing_key_prefix:
            if routing_key.startswith(routing_key_prefix):
                prefix_length = len(routing_key_prefix)
                return routing_key[prefix_length:]
        return routing_key

    @classmethod
    def decode_routing_key(cls, encoded_routing_key: str) -> str:
        def decode(match: Match) -> str:
            return binascii.unhexlify(match.group(1).encode("utf-8")).decode("utf-8")

        return re.sub(r"___([a-f0-9]{2}|[a-f0-9]{4}|[a-f0-9]{6}|[a-f0-9]{8})_", decode, encoded_routing_key)

    @classmethod
    def encode_routing_key(cls, routing_key: str) -> str:
        def encode(match: Match) -> str:
            return "___" + binascii.hexlify(match.group(1).encode("utf-8")).decode("utf-8") + "_"

        return re.sub(r'([^a-zA-Z0-9_*#,;.:<>!"%&/\(\)\[\]\{\}\\=?\'^`~+|@$ -])', encode, routing_key)

    @classmethod
    def get_queue_name(
        cls, routing_key: str, func_name: str, _uuid: str, competing_consumer: Optional[bool], context: Dict
    ) -> str:
        if not competing_consumer:
            queue_name = hashlib.sha256("{}{}{}".format(routing_key, func_name, _uuid).encode("utf-8")).hexdigest()
        else:
            queue_name = hashlib.sha256(routing_key.encode("utf-8")).hexdigest()

        queue_name_prefix: Optional[str] = cls.options(context).amqp.queue_name_prefix
        if queue_name_prefix:
            return "{}{}".format(queue_name_prefix, queue_name)
        return queue_name

    @classmethod
    def prefix_queue_name(cls, queue_name: str, context: Dict) -> str:
        queue_name_prefix: Optional[str] = cls.options(context).amqp.queue_name_prefix
        if queue_name_prefix:
            return "{}{}".format(queue_name_prefix, queue_name)
        return queue_name

    @classmethod
    async def subscribe_handler(
        cls,
        obj: Any,
        context: Dict,
        func: Any,
        routing_key: str,
        callback_kwargs: Optional[Union[List, Set, Tuple]] = None,
        exchange_name: str = "",
        competing: Optional[bool] = None,
        queue_name: Optional[str] = None,
        *,
        message_envelope: Any = MESSAGE_ENVELOPE_DEFAULT,
        message_protocol: Any = MESSAGE_ENVELOPE_DEFAULT,  # deprecated
        **kwargs: Any,
    ) -> Any:
        parser_kwargs = kwargs

        if message_envelope == MESSAGE_ENVELOPE_DEFAULT and message_protocol != MESSAGE_ENVELOPE_DEFAULT:
            # Fallback if deprecated message_protocol keyword is used
            message_envelope = message_protocol

        message_envelope = (
            context.get("message_envelope", context.get("message_protocol"))
            if message_envelope == MESSAGE_ENVELOPE_DEFAULT
            else message_envelope
        )

        # Validate the parser kwargs if there is a validation function in the envelope
        if message_envelope:
            envelope_kwargs_validation_func = getattr(message_envelope, "validate", None)
            if envelope_kwargs_validation_func:
                envelope_kwargs_validation_func(**parser_kwargs)

        _callback_kwargs: Any = callback_kwargs
        values = inspect.getfullargspec(func)
        if not _callback_kwargs:
            _callback_kwargs = (
                {
                    k: (
                        values.defaults[i - len(values.args) + 1]
                        if values.defaults and i >= len(values.args) - len(values.defaults) - 1
                        else None
                    )
                    for i, k in enumerate(values.args[1:])
                }
                if values.args and len(values.args) > 1
                else {}
            )
        else:
            _callback_kwargs = {k: None for k in _callback_kwargs if k != "self"}
        original_kwargs: Dict[str, Any] = {k: v for k, v in _callback_kwargs.items()}
        args_set = (set(values.args[1:]) | set(values.kwonlyargs) | set(callback_kwargs or [])) - set(["self"])

        async def handler(
            payload: Any, delivery_tag: Any, routing_key: str, properties: aioamqp.properties.Properties
        ) -> Any:
            logging.bind_logger(logging.getLogger("tomodachi.amqp").new(logger="tomodachi.amqp"))

            kwargs = dict(original_kwargs)

            message = payload
            message_uuid = None
            message_key = None
            if message_envelope:
                try:
                    parse_message_func = getattr(message_envelope, "parse_message", None)
                    if parse_message_func:
                        if len(parser_kwargs):
                            message, message_uuid, timestamp = await asyncio.create_task(
                                parse_message_func(payload, **parser_kwargs)
                            )
                        else:
                            message, message_uuid, timestamp = await asyncio.create_task(parse_message_func(payload))
                    if message_uuid:
                        if not context.get("_amqp_received_messages"):
                            context["_amqp_received_messages"] = {}
                        message_key = "{}:{}".format(message_uuid, func.__name__)
                        if context["_amqp_received_messages"].get(message_key):
                            return
                        context["_amqp_received_messages"][message_key] = time.time()
                        _received_messages = context["_amqp_received_messages"]
                        if (
                            _received_messages
                            and isinstance(_received_messages, dict)
                            and len(_received_messages) > 100000
                        ):
                            context["_amqp_received_messages"] = {
                                k: v for k, v in context["_amqp_received_messages"].items() if v > time.time() - 60
                            }

                    if args_set:
                        for k, v in message.items():
                            if k in args_set:
                                kwargs[k] = v
                        if "message" in args_set and (not isinstance(message, dict) or "message" not in message):
                            kwargs["message"] = message
                        if "routing_key" in args_set and (
                            not isinstance(message, dict) or "routing_key" not in message
                        ):
                            kwargs["routing_key"] = routing_key
                        if "exchange_name" in args_set and (
                            not isinstance(message, dict) or "exchange_name" not in message
                        ):
                            kwargs["exchange_name"] = exchange_name
                        if "properties" in args_set and (not isinstance(message, dict) or "properties" not in message):
                            kwargs["properties"] = properties
                        if "message_uuid" in args_set and (
                            not isinstance(message, dict) or "message_uuid" not in message
                        ):
                            kwargs["message_uuid"] = message_uuid
                except (Exception, asyncio.CancelledError, BaseException) as e:
                    limit_exception_traceback(e, ("tomodachi.transport.amqp",))
                    logging.getLogger("exception").exception("uncaught exception: {}".format(str(e)))
                    if message is not False and not message_uuid:
                        await cls.channel.basic_client_ack(delivery_tag)
                    elif message is False and message_uuid:
                        pass  # incompatible envelope, should probably ack if old message
                    elif message is False:
                        await cls.channel.basic_client_ack(delivery_tag)
                    return
            else:
                if args_set:
                    if "message" in args_set:
                        kwargs["message"] = message
                    if "routing_key" in args_set:
                        kwargs["routing_key"] = routing_key
                    if "exchange_name" in args_set:
                        kwargs["exchange_name"] = exchange_name
                    if "properties" in args_set:
                        kwargs["properties"] = properties
                    if "message_uuid" in args_set:
                        kwargs["message_uuid"] = message_uuid

                if len(values.args[1:]) and values.args[1] in kwargs:
                    del kwargs[values.args[1]]

            @functools.wraps(func)
            async def routine_func(*a: Any, **kw: Any) -> Any:
                logging.bind_logger(
                    logging.getLogger("tomodachi.amqp.handler").bind(handler=func.__name__, type="tomodachi.amqp")
                )
                get_contextvar("service.logger").set("tomodachi.amqp.handler")

                kw_values = {k: v for k, v in {**kwargs, **kw}.items() if values.varkw or k in args_set}
                args_values = [
                    kw_values.pop(key) if key in kw_values else a[i + 1]
                    for i, key in enumerate(values.args[1 : len(a) + 1])
                ]
                if values.varargs and not values.defaults and len(a) > len(args_values) + 1:
                    args_values += a[len(args_values) + 1 :]

                routine = func(*(obj, *args_values), **kw_values)
                if inspect.isawaitable(routine):
                    return_value = await routine
                else:
                    return_value = routine

                await cls.channel.basic_client_ack(delivery_tag)
                return return_value

            increase_execution_context_value("amqp_current_tasks")
            increase_execution_context_value("amqp_total_tasks")
            try:
                logging.bind_logger(
                    logging.getLogger("tomodachi.amqp.middleware").bind(
                        middleware=Ellipsis, handler=func.__name__, type="tomodachi.amqp"
                    )
                )
                return_value = await asyncio.create_task(
                    execute_middlewares(
                        func,
                        routine_func,
                        context.get("_amqp_message_pre_middleware", []) + context.get("message_middleware", []),
                        *(obj, message, routing_key),
                        message=message,
                        message_uuid=message_uuid,
                        routing_key=routing_key,
                        exchange_name=exchange_name,
                        properties=properties,
                    )
                )
            except (Exception, asyncio.CancelledError, BaseException) as e:
                limit_exception_traceback(
                    e,
                    ("tomodachi.transport.amqp", "tomodachi.helpers.middleware"),
                )
                logging.getLogger("exception").exception("uncaught exception: {}".format(str(e)))
                return_value = None
                if issubclass(
                    e.__class__,
                    (AmqpInternalServiceError, AmqpInternalServiceErrorException, AmqpInternalServiceException),
                ):
                    if message_key:
                        del context["_amqp_received_messages"][message_key]
                    await cls.channel.basic_client_nack(delivery_tag)
                else:
                    await cls.channel.basic_client_ack(delivery_tag)
            decrease_execution_context_value("amqp_current_tasks")

            return return_value

        exchange_name = exchange_name or cls.options(context).amqp.exchange_name
        context["_amqp_subscribers"] = context.get("_amqp_subscribers", [])
        context["_amqp_subscribers"].append((routing_key, exchange_name, competing, queue_name, func, handler))

        start_func = cls.subscribe(obj, context)
        return (await start_func) if start_func else None

    @classmethod
    async def connect(cls, obj: Any, context: Dict) -> Any:
        logger = logging.getLogger("tomodachi.amqp")
        logging.bind_logger(logger)

        logging.getLogger("aioamqp.protocol").setLevel(logging.WARNING)
        logging.getLogger("aioamqp.channel").setLevel(logging.WARNING)

        amqp_options: Options.AMQP = cls.options(context).amqp

        host = amqp_options.host
        port = amqp_options.port
        login = amqp_options.login
        password = amqp_options.password
        virtualhost = amqp_options.virtualhost
        ssl = amqp_options.ssl
        heartbeat = amqp_options.heartbeat

        try:
            protocol: Any
            transport, protocol = await aioamqp.connect(
                host=host,
                port=port,
                login=login,
                password=password,
                virtualhost=virtualhost,
                ssl=ssl,
                heartbeat=heartbeat,
            )
            cls.protocol = protocol
            cls.transport = transport
        except ConnectionRefusedError as e:
            error_message = "connection refused"
            logger.warning("Unable to connect [amqp] to {}:{} ({})".format(host, port, error_message))
            raise AmqpConnectionException(str(e), log_level=context.get("log_level")) from e
        except aioamqp.exceptions.AmqpClosedConnection as e:
            error_message = e.__context__
            logger.warning("Unable to connect [amqp] to {}:{} ({})".format(host, port, error_message))
            raise AmqpConnectionException(str(e), log_level=context.get("log_level")) from e
        except OSError as e:
            error_message = e.strerror if e.strerror else str(e)
            logger.warning("Unable to connect [amqp] to {}:{} ({})".format(host, port, error_message))
            raise AmqpConnectionException(str(e), log_level=context.get("log_level")) from e

        channel = await protocol.channel()
        if not cls.channel:
            stop_method = getattr(obj, "_stop_service", None)

            async def stop_service(*args: Any, **kwargs: Any) -> None:
                logging.getLogger("aioamqp.protocol").setLevel(logging.ERROR)
                await cls.protocol.close()
                cls.transport.close()
                cls.channel = None
                cls.transport = None
                cls.protocol = None
                if stop_method:
                    await stop_method(*args, **kwargs)

            setattr(obj, "_stop_service", stop_service)

        cls.channel = channel
        cls.exchange_name = cls.options(context).amqp.exchange_name

        return channel

    @classmethod
    async def subscribe(cls, obj: Any, context: Dict) -> Optional[Callable]:
        if context.get("_amqp_subscribed"):
            return None
        context["_amqp_subscribed"] = True

        logger = logging.getLogger("tomodachi.amqp")
        logging.bind_logger(logger)

        set_execution_context(
            {
                "amqp_enabled": True,
                "amqp_current_tasks": 0,
                "amqp_total_tasks": 0,
                "aioamqp_version": aioamqp.__version__,
            }
        )

        cls.channel = None
        channel = await cls.connect(obj, context)
        options: Options = cls.options(context)
        await channel.basic_qos(
            prefetch_count=options.amqp.qos.queue_prefetch_count, prefetch_size=0, connection_global=False
        )
        await channel.basic_qos(
            prefetch_count=options.amqp.qos.global_prefetch_count, prefetch_size=0, connection_global=True
        )

        async def _subscribe() -> None:
            logger = logging.getLogger("tomodachi.amqp")
            logging.bind_logger(logger)

            async def declare_queue(
                routing_key: str,
                func: Callable,
                exchange_name: str = "",
                exchange_type: str = "topic",
                queue_name: Optional[str] = None,
                passive: bool = False,
                durable: bool = True,
                exclusive: bool = False,
                auto_delete: bool = False,
                competing_consumer: Optional[bool] = None,
            ) -> Optional[str]:
                logger = logging.getLogger("tomodachi.amqp").bind(
                    wrapped_handler=func.__name__,
                    routing_key=routing_key or Ellipsis,
                    exchange_name=exchange_name or Ellipsis,
                    queue_name=queue_name or Ellipsis,
                )
                logging.bind_logger(logger)

                try:
                    if exchange_name and exchange_name != "amq.topic":
                        await channel.exchange_declare(
                            exchange_name=exchange_name,
                            type_name=exchange_type,
                            passive=False,
                            durable=True,
                            auto_delete=False,
                        )
                except aioamqp.exceptions.ChannelClosed as e:
                    error_message = e.args[1]
                    if e.args[0] == 403 and exchange_name.startswith("amq."):
                        logger.warning(
                            'Unable to declare exchange [amqp] "{}", starts with reserved "amq." ({})'.format(
                                exchange_name, error_message
                            )
                        )
                        raise
                    elif e.args[0] == 507 or e.args[0] == 406:
                        logger.warning(
                            'Unable to change type of existing exchange [amqp] "{}" ({})'.format(
                                exchange_name, error_message
                            )
                        )
                        raise
                    else:
                        logger.warning(
                            'Unable to declare exchange [amqp] "{}" ({})'.format(exchange_name, error_message)
                        )
                        raise

                if queue_name and competing_consumer is None:
                    competing_consumer = True

                _uuid = obj.uuid
                max_consumers = 1 if not competing_consumer else None

                if queue_name is None:
                    queue_name = cls.get_queue_name(
                        cls.encode_routing_key(routing_key), func.__name__, _uuid, competing_consumer, context
                    )
                else:
                    queue_name = cls.prefix_queue_name(queue_name, context)

                amqp_arguments = {}
                ttl = options.amqp.queue_ttl
                if ttl:
                    amqp_arguments["x-expires"] = int(ttl * 1000)

                try:
                    data = await channel.queue_declare(
                        queue_name,
                        passive=passive,
                        durable=durable,
                        exclusive=exclusive,
                        auto_delete=auto_delete,
                        arguments=amqp_arguments,
                    )
                    if max_consumers is not None and data.get("consumer_count", 0) >= max_consumers:
                        logger.warning(
                            'Max consumers ({}) for queue [amqp] "{}" has been reached'.format(
                                max_consumers, queue_name
                            )
                        )
                        raise AmqpTooManyConsumersException("Max consumers for this queue has been reached")
                except aioamqp.exceptions.ChannelClosed as e:
                    if e.args[0] == 405:
                        raise AmqpExclusiveQueueLockedException(str(e)) from e
                    raise AmqpException(str(e)) from e

                await channel.queue_bind(
                    queue_name,
                    exchange_name or "amq.topic",
                    cls.encode_routing_key(cls.get_routing_key(routing_key, context)),
                )

                return queue_name

            def callback(routing_key: str, handler: Callable) -> Callable:
                async def _callback(
                    self: Any, body: bytes, envelope: Any, properties: aioamqp.properties.Properties
                ) -> None:
                    # await channel.basic_reject(delivery_tag, requeue=True)
                    await asyncio.shield(handler(body.decode(), envelope.delivery_tag, routing_key, properties))

                return _callback

            for routing_key, exchange_name, competing, queue_name, func, handler in context.get(
                "_amqp_subscribers", []
            ):
                queue_name = await asyncio.create_task(
                    declare_queue(
                        routing_key,
                        func,
                        exchange_name=exchange_name,
                        competing_consumer=competing,
                        queue_name=queue_name,
                    )
                )
                await channel.basic_consume(callback(routing_key, handler), queue_name=queue_name)

        return _subscribe


__amqp = AmqpTransport.decorator(AmqpTransport.subscribe_handler)

amqp_publish = AmqpTransport.publish
publish = AmqpTransport.publish


def amqp(
    routing_key: str,
    callback_kwargs: Optional[Union[List, Set, Tuple]] = None,
    exchange_name: str = "",
    competing: Optional[bool] = None,
    queue_name: Optional[str] = None,
    *,
    message_envelope: Any = MESSAGE_ENVELOPE_DEFAULT,
    message_protocol: Any = MESSAGE_ENVELOPE_DEFAULT,  # deprecated
    **kwargs: Any,
) -> Callable:
    return cast(
        Callable,
        __amqp(
            routing_key,
            callback_kwargs=callback_kwargs,
            exchange_name=exchange_name,
            competing=competing,
            queue_name=queue_name,
            message_envelope=message_envelope,
            message_protocol=message_protocol,
            **kwargs,
        ),
    )



from __future__ import annotations

import logging
from traceback import format_exception
from typing import TYPE_CHECKING, Any, Dict, Optional, cast

from opentelemetry import trace
from opentelemetry._logs import get_logger_provider
from opentelemetry.sdk._logs import LoggerProvider, LoggingHandler
from opentelemetry.util.types import Attributes

from tomodachi.__version__ import __version__ as tomodachi_version

if TYPE_CHECKING:  # pragma: no cover
    try:
        from structlog.typing import EventDict, WrappedLogger
    except (ImportError, ModuleNotFoundError):
        from structlog.types import EventDict, WrappedLogger


def add_trace_structlog_processor(logger: WrappedLogger, method_name: str, event_dict: EventDict) -> EventDict:
    span = trace.get_current_span()
    ctx = span.get_span_context()

    if not ctx.is_valid:
        return event_dict

    event_dict["span_id"] = hex(ctx.span_id)
    event_dict["trace_id"] = hex(ctx.trace_id)

    parent: Optional[trace.SpanContext] = getattr(span, "parent", None)
    if parent:
        event_dict["parent_span_id"] = hex(parent.span_id)

    return event_dict


class OpenTelemetryLoggingHandler(LoggingHandler):
    def __init__(self, level: int = logging.NOTSET, logger_provider: Optional[LoggerProvider] = None) -> None:
        super().__init__(level=level)
        self._logger_provider = cast(LoggerProvider, logger_provider or get_logger_provider())
        self._logger = self._logger_provider.get_logger("tomodachi.opentelemetry", tomodachi_version)

    @staticmethod
    def _get_attributes(record: logging.LogRecord) -> Attributes:
        attributes = cast(Dict[str, Any], LoggingHandler._get_attributes(record))
        if record.exc_info:
            _, exc, tb = record.exc_info
            if exc is not None and exc.__traceback__ and exc.__traceback__ != tb:
                attributes["exception.stacktrace"] = "".join(format_exception(type(exc), exc, exc.__traceback__))
        logger_context: Dict[str, Any] = attributes.pop("logger.context", None)
        if not logger_context:
            return attributes
        for k, v in logger_context.items():
            if not isinstance(v, (bool, str, bytes, int, float)):
                attributes[f"{k}"] = str(v)
            else:
                attributes[f"{k}"] = v
        return attributes

    def emit(self, record: logging.LogRecord) -> None:
        record_ = self._translate(record)
        self._logger.emit(record_)



from tomodachi.envelope import json_base as json_base
from tomodachi.envelope import protobuf_base as protobuf_base
from tomodachi.envelope.json_base import JsonBase as JsonBase
from tomodachi.envelope.protobuf_base import ProtobufBase as ProtobufBase



import asyncio
import functools
import inspect
from typing import Any, Callable, Dict, List, Set, Tuple, cast

from tomodachi import get_contextvar, logging

TOMODACHI_MIDDLEWARE_ATTRIBUTE = "_tomodachi_middleware_attribute"


async def execute_middlewares(
    func: Callable, routine_func: Callable, middlewares: List, *args: Any, **init_kwargs: Any
) -> Any:
    if middlewares:
        logger = logging.getLogger()

        middleware_context: Dict = {}

        async def middleware_wrapper(idx: int = 0, *ma: Any, **mkw: Any) -> Any:
            middleware: Callable = middlewares[idx]

            logging.bind_logger(
                logger.bind(
                    middleware=(
                        getattr(middleware, "name", Ellipsis)
                        if hasattr(middleware, "name")
                        else (
                            middleware.__name__
                            if hasattr(middleware, "__name__")
                            else (
                                type(middleware).__name__
                                if hasattr(type(middleware), "__name__")
                                else str(type(middleware))
                            )
                        )
                    )
                )
            )
            get_contextvar("service.logger").set(logger._context["logger"])

            if getattr(middleware, TOMODACHI_MIDDLEWARE_ATTRIBUTE, None) is None:
                values = inspect.getfullargspec(inspect.unwrap(middleware))
                middleware_kwargs = set(values.args + values.kwonlyargs)
                middleware_args = values.args
                has_defaults = True if values.defaults else False
                has_varargs = True if values.varargs else False
                has_varkw = True if values.varkw else False

                is_bound_method = bool(
                    inspect.ismethod(middleware)
                    or getattr(middleware, "__self__", None) is middleware
                    or inspect.ismethod(getattr(middleware, "__call__", None))
                )
                arg_start = 0 if is_bound_method else 1
                arg_len = len(values.args) - len(values.defaults or ()) + arg_start
                middleware_args_offset = 2 if is_bound_method else 1

                setattr(
                    middleware,
                    TOMODACHI_MIDDLEWARE_ATTRIBUTE,
                    (
                        arg_len,
                        arg_start,
                        middleware_kwargs,
                        middleware_args,
                        has_defaults,
                        has_varargs,
                        has_varkw,
                        is_bound_method,
                        middleware_args_offset,
                    ),
                )
            else:
                (
                    arg_len,
                    arg_start,
                    middleware_kwargs,
                    middleware_args,
                    has_defaults,
                    has_varargs,
                    has_varkw,
                    is_bound_method,
                    middleware_args_offset,
                ) = cast(
                    Tuple[int, int, Set[str], List[str], bool, bool, bool, bool, int],
                    getattr(middleware, TOMODACHI_MIDDLEWARE_ATTRIBUTE),
                )

            if has_varargs and not has_defaults:
                arg_len = 3 + len(args)

            if middlewares and len(middlewares) <= idx + 1:

                @functools.wraps(func)
                async def _func_wrapper(*a: Any, **kw: Any) -> Any:
                    return await asyncio.create_task(routine_func(*args, **{**mkw, **kw, **init_kwargs}))

                middleware_arguments = [middleware, _func_wrapper, *args, middleware_context][arg_start:arg_len]
            else:

                @functools.wraps(func)
                async def _middleware_wrapper(*a: Any, **kw: Any) -> Any:
                    return await asyncio.create_task(middleware_wrapper(idx + 1, *a, **{**mkw, **kw, **init_kwargs}))

                middleware_arguments = [middleware, _middleware_wrapper, *args, middleware_context][arg_start:arg_len]

            mkw_cleaned = {k: v for k, v in mkw.items() if has_varkw or k in middleware_kwargs}

            for i, key in enumerate(
                middleware_args[middleware_args_offset : len(middleware_arguments)], middleware_args_offset
            ):
                if key in mkw_cleaned:
                    middleware_arguments[i] = mkw_cleaned.pop(key)

            if is_bound_method:
                return await asyncio.create_task(middleware(*middleware_arguments[1:], **mkw_cleaned))

            return await asyncio.create_task(middleware(*middleware_arguments, **mkw_cleaned))

        return_value = await asyncio.create_task(middleware_wrapper(**init_kwargs))
    else:
        return_value = await asyncio.create_task(routine_func(*args, **init_kwargs))

    return return_value



import os
import sys

if sys.argv and sys.argv[0].rsplit(os.sep, 1)[-1] == "opentelemetry-instrument":
    # this is a hacky way of setting the default distro and configurator used during auto instrumentation.
    # see ".../site-packages/opentelemetry/instrumentation/auto_instrumentation/__init__.py"
    for key in ("OTEL_PYTHON_DISTRO", "OTEL_PYTHON_CONFIGURATOR"):
        if not os.environ.get(key):
            os.environ.setdefault(key, "tomodachi")


OTEL_PYTHON_TOMODACHI_EXCLUDED_URLS = "OTEL_PYTHON_TOMODACHI_EXCLUDED_URLS"
"""
.. envvar:: OTEL_PYTHON_TOMODACHI_EXCLUDED_URLS

The :envvar:`OTEL_PYTHON_TOMODACHI_EXCLUDED_URLS` is used to define regex patterns of HTTP URLs that should be
excluded from instrumentation.
Default: ""
"""

OTEL_PYTHON_TOMODACHI_PROMETHEUS_ADDRESS = "OTEL_PYTHON_TOMODACHI_PROMETHEUS_ADDRESS"
"""
.. envvar:: OTEL_PYTHON_TOMODACHI_PROMETHEUS_ADDRESS

The :envvar:`OTEL_PYTHON_TOMODACHI_PROMETHEUS_ADDRESS` specifies the address on which the
"tomodachi_prometheus" meter provider will listen on.
Default: "localhost"
"""

OTEL_PYTHON_TOMODACHI_PROMETHEUS_PORT = "OTEL_PYTHON_TOMODACHI_PROMETHEUS_PORT"
"""
.. envvar:: OTEL_PYTHON_TOMODACHI_PROMETHEUS_PORT

The :envvar:`OTEL_PYTHON_TOMODACHI_PROMETHEUS_PORT` specifies the port on which the
"tomodachi_prometheus" meter provider will listen on.
Default: 9464
"""

OTEL_PYTHON_TOMODACHI_PROMETHEUS_NAMESPACE_PREFIX = "OTEL_PYTHON_TOMODACHI_PROMETHEUS_NAMESPACE_PREFIX"
"""
.. envvar:: OTEL_PYTHON_TOMODACHI_PROMETHEUS_NAMESPACE_PREFIX

The :envvar:`OTEL_PYTHON_TOMODACHI_PROMETHEUS_NAMESPACE_PREFIX` configures the exporter to prefix metrics
with the given namespace. Metadata metrics such as target_info and otel_scope_info are not prefixed since
these have special behavior based on their name.
Default: ""
"""

OTEL_PYTHON_TOMODACHI_PROMETHEUS_INCLUDE_SCOPE_INFO = "OTEL_PYTHON_TOMODACHI_PROMETHEUS_INCLUDE_SCOPE_INFO"
"""
.. envvar:: OTEL_PYTHON_TOMODACHI_PROMETHEUS_INCLUDE_SCOPE_INFO

The :envvar:`OTEL_PYTHON_TOMODACHI_PROMETHEUS_INCLUDE_SCOPE_INFO` if set to True will export a Prometheus
info metric "otel_scope_info" with labels "otel_scope_name" and "otel_scope_version", as well as the labels
are also added to all metric points. Additional details are available `in the specification
<https://opentelemetry.io/docs/specs/otel/compatibility/prometheus_and_openmetrics/#instrumentation-scope-1`__.
Default: True
"""

OTEL_PYTHON_TOMODACHI_PROMETHEUS_INCLUDE_TARGET_INFO = "OTEL_PYTHON_TOMODACHI_PROMETHEUS_INCLUDE_TARGET_INFO"
"""
.. envvar:: OTEL_PYTHON_TOMODACHI_PROMETHEUS_INCLUDE_TARGET_INFO

The :envvar:`OTEL_PYTHON_TOMODACHI_PROMETHEUS_INCLUDE_TARGET_INFO` if set to True will export a Prometheus
info metric "target_info" which includes the labels "job" and "instance", describing the service.namespace,
service.name and service.instance.id resource attributes. These labels will also be added to all metric points.
Additional details are available `in the specification
<https://opentelemetry.io/docs/specs/otel/compatibility/prometheus_and_openmetrics/#resource-attributes-1`__.
Default: True
"""

OTEL_PYTHON_TOMODACHI_PROMETHEUS_EXEMPLARS_ENABLED = "OTEL_PYTHON_TOMODACHI_PROMETHEUS_EXEMPLARS_ENABLED"
"""
.. envvar:: OTEL_PYTHON_TOMODACHI_PROMETHEUS_EXEMPLARS_ENABLED

The :envvar:`OTEL_PYTHON_TOMODACHI_PROMETHEUS_EXEMPLARS_ENABLED` if set to True will include exemplars in
the Prometheus output when queried with the OpenMetrics accept header. Use of these exemplars with the
Prometheus meter provider are currently experimental.
Default: False
"""



from typing import (
    Any,
    Dict,
    ItemsView,
    Iterator,
    KeysView,
    Literal,
    Optional,
    Set,
    TypedDict,
    TypeVar,
    Union,
    ValuesView,
    cast,
    overload,
)

from botocore.credentials import Credentials as BotocoreCredentials
from botocore.credentials import ReadOnlyCredentials as BotocoreReadOnlyCredentials

T = TypeVar("T")


class CredentialsDict(TypedDict, total=False):
    region_name: Optional[str]
    aws_access_key_id: Optional[str]
    aws_secret_access_key: Optional[str]
    aws_session_token: Optional[str]
    endpoint_url: Optional[str]


class CredentialsTypeProtocol:
    region_name: Optional[str]
    aws_access_key_id: Optional[str]
    aws_secret_access_key: Optional[str]
    aws_session_token: Optional[str]
    endpoint_url: Optional[str]


CredentialsTypeKeys = Literal[
    "region_name", "aws_access_key_id", "aws_secret_access_key", "aws_session_token", "endpoint_url"
]


class Credentials:
    region_name: Optional[str]
    aws_access_key_id: Optional[str]
    aws_secret_access_key: Optional[str]
    aws_session_token: Optional[str]
    endpoint_url: Optional[str]
    __unset: Set[str]

    def keys(
        self,
    ) -> KeysView[CredentialsTypeKeys]:
        result = {}
        for key in ("region_name", "aws_access_key_id", "aws_secret_access_key", "aws_session_token", "endpoint_url"):
            result[key] = ...
        return cast(Dict[CredentialsTypeKeys, Optional[str]], result).keys()

    def dict(self) -> CredentialsDict:
        result: CredentialsDict = {}
        for key in self.keys():
            value = getattr(self, key, None)
            if value is Ellipsis:
                value = None
            result[key] = value
        return result

    def values(
        self,
    ) -> ValuesView[Optional[str]]:
        return cast(Dict[CredentialsTypeKeys, Optional[str]], self.dict()).values()

    def items(
        self,
    ) -> ItemsView[CredentialsTypeKeys, Optional[str]]:
        return cast(Dict[CredentialsTypeKeys, Optional[str]], self.dict()).items()

    @overload
    def __init__(
        self,
        __map: Union[
            CredentialsDict,
            CredentialsTypeProtocol,
            "Credentials",
            BotocoreCredentials,
            BotocoreReadOnlyCredentials,
            Dict[
                Literal[
                    "region_name", "aws_access_key_id", "aws_secret_access_key", "aws_session_token", "endpoint_url"
                ],
                Optional[str],
            ],
        ],
        /,
        *,
        region_name: Optional[str] = ...,
        aws_access_key_id: Optional[str] = ...,
        aws_secret_access_key: Optional[str] = ...,
        aws_session_token: Optional[str] = ...,
        endpoint_url: Optional[str] = ...,
    ) -> None: ...

    @overload
    def __init__(
        self,
        /,
        *,
        region_name: Optional[str] = ...,
        aws_access_key_id: Optional[str] = ...,
        aws_secret_access_key: Optional[str] = ...,
        aws_session_token: Optional[str] = ...,
        endpoint_url: Optional[str] = ...,
    ) -> None: ...

    def __init__(
        self,
        __map: Optional[
            Union[
                CredentialsDict,
                CredentialsTypeProtocol,
                "Credentials",
                BotocoreCredentials,
                BotocoreReadOnlyCredentials,
                Dict[
                    Literal[
                        "region_name", "aws_access_key_id", "aws_secret_access_key", "aws_session_token", "endpoint_url"
                    ],
                    Optional[str],
                ],
            ]
        ] = None,
        /,
        **kwargs: Any,
    ) -> None:
        if __map is not None:
            if isinstance(__map, BotocoreCredentials):
                __map = __map.get_frozen_credentials()
            if isinstance(__map, BotocoreReadOnlyCredentials):
                __map = {
                    "aws_access_key_id": __map.access_key,
                    "aws_secret_access_key": __map.secret_key,
                    "aws_session_token": __map.token,
                }
            if not isinstance(__map, dict):
                try:
                    __map = cast(CredentialsDict, dict(__map))  # type: ignore[arg-type]
                except TypeError:
                    __map = cast(CredentialsDict, dict(__map.__dict__))
            if __map and isinstance(__map, dict):
                for key, value in __map.items():
                    setattr(self, key, value)
        for key in kwargs:
            if key not in self.keys():
                raise TypeError(f"__init__() got an unexpected keyword argument '{key}'")
            setattr(self, key, kwargs[key])

        self.__unset = set()
        for key in self.keys():
            try:
                self.get(key, default=Ellipsis)
            except AttributeError:
                setattr(self, key, None)
                self.__unset.add(key)

    def __iter__(
        self,
    ) -> Iterator[CredentialsTypeKeys]:
        return cast(Iterator[CredentialsTypeKeys], iter(cast(CredentialsDict, self.dict())))

    def __getitem__(
        self,
        key: CredentialsTypeKeys,
    ) -> Optional[str]:
        return self.get(key, default=Ellipsis) if key not in self.__unset else None

    def __contains__(
        self,
        key: str,
    ) -> bool:
        return key in self.keys()

    @overload
    def get(
        self,
        key: CredentialsTypeKeys,
        default: Any = ...,
    ) -> Optional[str]: ...

    @overload
    def get(
        self,
        key: str,
        default: T,
    ) -> T: ...

    def get(
        self,
        key: str,
        default: Any = None,
    ) -> Any:
        if default is Ellipsis:
            return getattr(self, key) if key not in self.__unset else None
        return getattr(self, key, default) if key not in self.__unset else default


CredentialsMapping = Union[
    Credentials, CredentialsDict, CredentialsTypeProtocol, BotocoreCredentials, BotocoreReadOnlyCredentials
]



from .base import FUNCTION_ATTRIBUTE, INVOKER_TASK_START_KEYWORD, START_ATTRIBUTE, Invoker  # noqa
from .decorator import decorator  # noqa

__all__ = [
    "FUNCTION_ATTRIBUTE",
    "INVOKER_TASK_START_KEYWORD",
    "START_ATTRIBUTE",
    "Invoker",
    "decorator",
]



import functools
import types
from typing import Any, Callable, Dict, Optional, Tuple, cast

from tomodachi.options import Options

FUNCTION_ATTRIBUTE = "_tomodachi_function_is_invoker_function"
START_ATTRIBUTE = "_tomodachi_deprecated_invoker_function_start_marker"
INVOKER_TASK_START_KEYWORD = "_tomodachi_invoker_task_start_keyword"


class Invoker(object):
    context: Dict = {}

    @classmethod
    def decorator(cls, cls_func: Callable) -> Callable:
        def _wrapper(*args: Any, **kwargs: Any) -> Callable:
            def wrapper(func: Callable) -> Callable:
                fn: Optional[Callable] = func
                unwrapped_func = func
                while fn:
                    unwrapped_func = fn
                    if getattr(fn, "__wrapped__", None) and not getattr(fn, FUNCTION_ATTRIBUTE, None):
                        break
                    fn = getattr(fn, "__wrapped__", None)

                @functools.wraps(unwrapped_func)
                async def _decorator(obj: Any, *a: Any, **kw: Any) -> Any:
                    if not kw or not kw.get(INVOKER_TASK_START_KEYWORD):
                        return await unwrapped_func(obj, *a, **kw)

                    setattr(_decorator, START_ATTRIBUTE, False)  # deprecated
                    if not cls.context.get(obj, None):
                        if getattr(obj, "context", None):
                            cls.context[obj] = obj.context
                        else:
                            cls.context[obj] = {}
                        cls.context[obj].update(
                            {
                                i: getattr(obj, i)
                                for i in dir(obj)
                                if not callable(i)
                                and not i.startswith("__")
                                and not isinstance(getattr(obj, i), types.MethodType)
                            }
                        )
                    context = cls.context[obj]
                    obj.context = context
                    bound_cls_func = getattr(cls_func, "__get__")(cls)
                    start_func = await bound_cls_func(obj, context, unwrapped_func, *args, **kwargs)

                    # Work-around if the decorators are stacked with multiple decorators for the same method
                    if getattr(func, FUNCTION_ATTRIBUTE, None):
                        decorated_func = cast(Callable, getattr(func, "func"))
                        decorated_cls_func = cast(Callable, getattr(func, "cls_func"))
                        decorated_args = cast(Tuple, getattr(func, "args"))
                        decorated_kwargs = cast(Dict, getattr(func, "kwargs"))

                        fn = cast(
                            Callable,
                            cls.decorator(decorated_cls_func)(*decorated_args, **decorated_kwargs)(decorated_func),
                        )
                        setattr(fn, START_ATTRIBUTE, True)
                        await fn(obj, *a, **kw)

                    return start_func

                # Work-around if the decorators are stacked with multiple decorators for the same method
                if not getattr(_decorator, FUNCTION_ATTRIBUTE, None):
                    setattr(_decorator, "func", func)
                    setattr(_decorator, "cls_func", cls_func)
                    setattr(_decorator, "args", args)
                    setattr(_decorator, "kwargs", kwargs)

                    setattr(_decorator, FUNCTION_ATTRIBUTE, True)

                return _decorator

            if not kwargs and len(args) == 1 and callable(args[0]):
                func = args[0]
                args = ()
                return wrapper(func)
            else:
                return wrapper

        return _wrapper

    @staticmethod
    def options(context: Dict) -> Options:
        return cast(
            Options,
            (
                context.get("options")
                if isinstance(context.get("options"), Options)
                else Options(**context.get("options", {}))
            ),
        )



import importlib
import importlib.util
import os
import sys
from types import ModuleType
from typing import Any

import tomodachi
from tomodachi import logging


class ServicePackageError(ImportError):
    pass


class ServiceImporter(object):
    @classmethod
    def import_service_file(cls, file_name: str, no_exec: bool = False) -> ModuleType:
        cwd = os.getcwd()
        file_path = "{}/{}.py".format(os.path.realpath(cwd), file_name)
        if file_path.endswith(".py.py"):
            file_path = file_path[:-3]
        try:
            sys.path.insert(0, cwd)
            sys.path.insert(0, os.path.dirname(os.path.dirname(file_path)))
            if file_path.endswith(".py") and not os.path.isfile(file_path):
                raise OSError("No such service file")
            if not file_path.endswith(".py") and not os.path.isfile("{}.py".format(file_path)):
                raise OSError("No such service file")
            try:
                spec: Any = importlib.util.find_spec(
                    ".{}".format(file_path.rsplit("/", 1)[1])[:-3], package=os.path.dirname(file_path).rsplit("/", 1)[1]
                )
                if not spec:
                    # package name already taken, imported service not part of a package
                    spec = importlib.util.spec_from_file_location(file_name, file_path)
            except AttributeError as e:
                # package name already taken, imported service not part of a package
                try:
                    spec = importlib.util.spec_from_file_location(file_name, file_path)
                except Exception:
                    raise e
            except ModuleNotFoundError as e:  # noqa
                file_path_package_name = file_path[:-3] if file_path.endswith(".py") else file_path
                if str(e) == "__path__ attribute not found on '{}' while trying to find '{}'".format(
                    file_path_package_name.rsplit("/", 2)[1], ".".join(file_path_package_name.rsplit("/", 2)[1:])
                ):
                    logging.getLogger("tomodachi.importer").warning(
                        "invalid service package name or parent name may conflict with builtin python modules - change parent folder name",
                        package_name=file_path.rsplit("/", 2)[1],
                    )
                    raise ServicePackageError from e
                if str(e) == "__path__ attribute not found on '{}'while trying to find '{}'".format(
                    file_path_package_name.rsplit("/", 2)[1], ".".join(file_path_package_name.rsplit("/", 2)[1:])
                ):
                    logging.getLogger("tomodachi.importer").warning(
                        "invalid service package name or parent name may conflict with builtin python modules - change parent folder name",
                        package_name=file_path.rsplit("/", 2)[1],
                    )
                    raise ServicePackageError from e
                raise e
            if not spec:
                raise OSError
            service_import = importlib.util.module_from_spec(spec)
            try:
                importlib.reload(service_import)
                service_import = importlib.util.module_from_spec(spec)
            except ImportError:
                pass
            if not spec.loader:
                raise OSError
            service_import_name = ""
            try:
                if service_import:
                    service_import_name = (
                        service_import.__name__[:-3]
                        if service_import.__name__.endswith(".py")
                        else service_import.__name__
                    )
                else:
                    service_import_name = ""
                if no_exec:
                    return service_import
                spec.loader.exec_module(service_import)
            except ImportError as e:
                if service_import_name and str(e) == "No module named '{}'".format(service_import_name):
                    logging.getLogger("tomodachi.importer").warning(
                        "invalid service package name or parent name may conflict with builtin python modules - change parent folder name",
                        package_name=file_path.rsplit("/", 2)[1],
                    )
                    raise ServicePackageError from e
                if str(e) == "attempted relative import with no known parent package":
                    logging.getLogger("tomodachi.importer").warning(
                        "invalid service package name or parent name may conflict with builtin python modules - change parent folder name",
                        package_name=file_path.rsplit("/", 2)[1],
                    )
                    raise ServicePackageError from e
                raise e
            except SystemError as e:
                if service_import_name and str(
                    e
                ) == "Parent module '{}' not loaded, cannot perform relative import".format(service_import_name):
                    logging.getLogger("tomodachi.importer").warning(
                        "invalid service package name or parent name may conflict with builtin python modules - change parent folder name",
                        package_name=file_path.rsplit("/", 2)[1],
                    )
                    raise ServicePackageError from e
                if str(e) == "Parent module '' not loaded, cannot perform relative import":
                    logging.getLogger("tomodachi.importer").warning(
                        "invalid service package name or parent name may conflict with builtin python modules - change parent folder name",
                        package_name=file_path.rsplit("/", 2)[1],
                    )
                    raise ServicePackageError from e
                raise e
        except (ImportError, ModuleNotFoundError) as e:  # noqa
            if file_name.endswith(".py.py"):
                return cls.import_service_file(file_name[:-3])
            if file_name.endswith(".py") and isinstance(e, ModuleNotFoundError):  # noqa
                return cls.import_service_file(file_name[:-3])
            if file_name.endswith(".py"):
                file_name = file_name[:-3]
            logging.getLogger("tomodachi.importer").warning(
                "unable to load service file", file_path="{}.py".format(file_name)
            )
            raise e
        except OSError:
            if file_name.endswith(".py"):
                file_name = file_name[:-3]
            logging.getLogger("tomodachi.importer").error("no such service file", file_path="{}.py".format(file_name))
            tomodachi.SERVICE_EXIT_CODE = 2
            raise
        except Exception as e:
            if file_name.endswith(".py"):
                file_name = file_name[:-3]
            logging.getLogger("tomodachi.importer").error(
                "unable to load service file", file_path="{}.py".format(file_name)
            )
            if not isinstance(e, (SyntaxError, IndentationError)):
                logging.getLogger("tomodachi.importer").warning("error during import", exception=e)
            raise e
        return service_import

    @classmethod
    def import_module(cls, file_name: str, no_exec: bool = False) -> ModuleType:
        cwd = os.getcwd()
        file_path = "{}/{}".format(os.path.realpath(cwd), file_name)

        spec: Any = importlib.util.spec_from_file_location(file_name, file_path)
        module_import = importlib.util.module_from_spec(spec)
        if no_exec:
            return module_import
        spec.loader.exec_module(module_import)

        return module_import



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: sns_sqs_message.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x15sns_sqs_message.proto\x12\x05proto\"%\n\x07Service\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0c\n\x04uuid\x18\x02 \x01(\t\"s\n\x08Metadata\x12\x14\n\x0cmessage_uuid\x18\x01 \x01(\t\x12\x18\n\x10protocol_version\x18\x02 \x01(\t\x12\x11\n\ttimestamp\x18\x03 \x01(\x01\x12\r\n\x05topic\x18\x04 \x01(\t\x12\x15\n\rdata_encoding\x18\x05 \x01(\t\"a\n\rSNSSQSMessage\x12\x1f\n\x07service\x18\x01 \x01(\x0b\x32\x0e.proto.Service\x12!\n\x08metadata\x18\x02 \x01(\x0b\x32\x0f.proto.Metadata\x12\x0c\n\x04\x64\x61ta\x18\x03 \x01(\x0c\x62\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'sns_sqs_message_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _SERVICE._serialized_start=32
  _SERVICE._serialized_end=69
  _METADATA._serialized_start=71
  _METADATA._serialized_end=186
  _SNSSQSMESSAGE._serialized_start=188
  _SNSSQSMESSAGE._serialized_end=285
# @@protoc_insertion_point(module_scope)



import functools
import logging
from os import environ
from typing import Any, Collection, Dict, List, Optional, Set, Union, cast

import structlog
from opentelemetry.instrumentation.instrumentor import BaseInstrumentor  # type: ignore
from opentelemetry.metrics import NoOpMeter, get_meter
from opentelemetry.sdk._logs import LoggerProvider
from opentelemetry.sdk.environment_variables import OTEL_LOG_LEVEL
from opentelemetry.sdk.metrics import Meter, MeterProvider
from opentelemetry.sdk.resources import SERVICE_INSTANCE_ID as RESOURCE_SERVICE_INSTANCE_ID
from opentelemetry.sdk.resources import SERVICE_NAME as RESOURCE_SERVICE_NAME
from opentelemetry.sdk.resources import Resource
from opentelemetry.sdk.trace import Tracer, TracerProvider
from opentelemetry.trace import NoOpTracer, SpanKind, StatusCode, get_tracer
from opentelemetry.trace.propagation.tracecontext import TraceContextTextMapPropagator
from opentelemetry.util.http import ExcludeList, get_excluded_urls, parse_excluded_urls
from opentelemetry.util.types import AttributeValue

import tomodachi
from tomodachi.__version__ import __version__ as tomodachi_version
from tomodachi.opentelemetry.distro import (
    _add_meter_provider_views,
    _create_logger_provider,
    _create_meter_provider,
    _create_tracer_provider,
    _get_logger_provider,
    _get_meter_provider,
    _get_tracer_provider,
)
from tomodachi.opentelemetry.logging import OpenTelemetryLoggingHandler, add_trace_structlog_processor
from tomodachi.opentelemetry.middleware import (
    OpenTelemetryAioHTTPMiddleware,
    OpenTelemetryAMQPMiddleware,
    OpenTelemetryAWSSQSMiddleware,
    OpenTelemetryScheduleFunctionMiddleware,
)

INSTRUMENTATION_CLASS_ATTRIBUTES = {
    "_opentelemetry_tracer_provider",
    "_opentelemetry_meter_provider",
    "_opentelemetry_excluded_urls",
    "_is_instrumented_by_opentelemetry",
    "_InstrumentedTomodachiService__post_init_hook",
    "_InstrumentedTomodachiService__post_teardown_hook",
}


class TomodachiInstrumentor(BaseInstrumentor):
    _instrumented_services: Optional[Set[tomodachi.Service]] = None
    _logging_handlers: Optional[List[OpenTelemetryLoggingHandler]] = None
    _is_instrumented_by_opentelemetry: bool = False

    @classmethod
    def instrument_service(
        cls,
        service: tomodachi.Service,
        tracer_provider: Optional[TracerProvider] = None,
        meter_provider: Optional[MeterProvider] = None,
        excluded_urls: Optional[str] = None,
    ) -> None:
        if isinstance(service, type):
            raise Exception("instrument_service must be called with an instance of tomodachi.Service")

        tracer_provider = cls._tracer_provider(tracer_provider)
        meter_provider = cls._meter_provider(meter_provider)

        if excluded_urls is None:
            _excluded_urls = ExcludeList(
                list(
                    set(
                        [url for url in get_excluded_urls("TOMODACHI")._excluded_urls]
                        + [url for url in get_excluded_urls("AIOHTTP")._excluded_urls]
                    )
                )
            )
        else:
            _excluded_urls = parse_excluded_urls(excluded_urls)

        if cls._instrumented_services is None:
            cls._instrumented_services = set()

        cls._instrumented_services.add(service)

        # setting resource "service.name" if not set - this is a bit hacky at the moment
        additional_resource_attributes: Dict[str, AttributeValue]
        if getattr(tracer_provider, "resource", None):
            additional_resource_attributes = {}
            if service.name not in ("service", "app"):
                attr_value = tracer_provider.resource._attributes.get(RESOURCE_SERVICE_NAME) or ""
                if attr_value[0:16] in ("", "unknown_service", "unknown_service:"):
                    additional_resource_attributes[RESOURCE_SERVICE_NAME] = service.name
                attr_value = tracer_provider.resource._attributes.get(RESOURCE_SERVICE_INSTANCE_ID) or ""
                if not attr_value:
                    additional_resource_attributes[RESOURCE_SERVICE_INSTANCE_ID] = service.uuid
            if additional_resource_attributes:
                resource = tracer_provider.resource.merge(
                    Resource.create({**tracer_provider.resource._attributes, **additional_resource_attributes})
                )
                tracer_provider.resource._attributes = resource._attributes

        if getattr(meter_provider._sdk_config, "resource", None):
            additional_resource_attributes = {}
            if service.name not in ("service", "app"):
                attr_value = meter_provider._sdk_config.resource._attributes.get(RESOURCE_SERVICE_NAME) or ""
                if attr_value[0:16] in ("", "unknown_service", "unknown_service:"):
                    additional_resource_attributes[RESOURCE_SERVICE_NAME] = service.name
            attr_value = meter_provider._sdk_config.resource._attributes.get(RESOURCE_SERVICE_INSTANCE_ID) or ""
            if not attr_value:
                additional_resource_attributes[RESOURCE_SERVICE_INSTANCE_ID] = service.uuid
            if additional_resource_attributes:
                resource = meter_provider._sdk_config.resource.merge(
                    Resource.create(
                        {**meter_provider._sdk_config.resource._attributes, **additional_resource_attributes}
                    )
                )
                meter_provider._sdk_config.resource._attributes = resource._attributes

        if cls._logging_handlers:
            for handler in cls._logging_handlers:
                logger_provider = cast(LoggerProvider, handler._logger_provider)
                if getattr(logger_provider, "resource", None):
                    additional_resource_attributes = {}
                    if service.name not in ("service", "app"):
                        attr_value = logger_provider.resource._attributes.get(RESOURCE_SERVICE_NAME) or ""
                        if attr_value[0:16] in ("", "unknown_service", "unknown_service:"):
                            additional_resource_attributes[RESOURCE_SERVICE_NAME] = service.name
                    attr_value = logger_provider.resource._attributes.get(RESOURCE_SERVICE_INSTANCE_ID) or ""
                    if not attr_value:
                        additional_resource_attributes[RESOURCE_SERVICE_INSTANCE_ID] = service.uuid
                    if additional_resource_attributes:
                        resource = logger_provider.resource.merge(
                            Resource.create({**logger_provider.resource._attributes, **additional_resource_attributes})
                        )
                        logger_provider.resource._attributes = resource._attributes
                        setattr(handler._logger, "_resource", logger_provider.resource)

        tracer = get_tracer("tomodachi", tomodachi_version, tracer_provider)
        meter = get_meter("tomodachi", tomodachi_version, meter_provider)

        if meter and not isinstance(meter, NoOpMeter) and not tracer:
            # collect metrics even if tracing is disabled
            tracer = NoOpTracer()

        aiohttp_pre_middleware = getattr(service, "_aiohttp_pre_middleware", None)
        if aiohttp_pre_middleware is None:
            aiohttp_pre_middleware = []
            setattr(service, "_aiohttp_pre_middleware", aiohttp_pre_middleware)
        if not [m for m in aiohttp_pre_middleware if isinstance(m, OpenTelemetryAioHTTPMiddleware)]:
            aiohttp_pre_middleware.append(
                OpenTelemetryAioHTTPMiddleware(
                    service=service, tracer=tracer, meter=meter, excluded_urls=_excluded_urls
                )
            )

        awssnssqs_message_pre_middleware = getattr(service, "_awssnssqs_message_pre_middleware", None)
        if awssnssqs_message_pre_middleware is None:
            awssnssqs_message_pre_middleware = []
            setattr(service, "_awssnssqs_message_pre_middleware", awssnssqs_message_pre_middleware)
        if not [m for m in awssnssqs_message_pre_middleware if isinstance(m, OpenTelemetryAWSSQSMiddleware)]:
            awssnssqs_message_pre_middleware.append(
                OpenTelemetryAWSSQSMiddleware(service=service, tracer=tracer, meter=meter)
            )

        amqp_message_pre_middleware = getattr(service, "_amqp_message_pre_middleware", None)
        if amqp_message_pre_middleware is None:
            amqp_message_pre_middleware = []
            setattr(service, "_amqp_message_pre_middleware", amqp_message_pre_middleware)
        if not [m for m in amqp_message_pre_middleware if isinstance(m, OpenTelemetryAMQPMiddleware)]:
            amqp_message_pre_middleware.append(OpenTelemetryAMQPMiddleware(service=service, tracer=tracer, meter=meter))

        schedule_pre_middleware = getattr(service, "_schedule_pre_middleware", None)
        if schedule_pre_middleware is None:
            schedule_pre_middleware = []
            setattr(service, "_schedule_pre_middleware", schedule_pre_middleware)
        if not [m for m in schedule_pre_middleware if isinstance(m, OpenTelemetryScheduleFunctionMiddleware)]:
            schedule_pre_middleware.append(
                OpenTelemetryScheduleFunctionMiddleware(service=service, tracer=tracer, meter=meter)
            )

        context = getattr(service, "context", None)
        if context:
            context["_aiohttp_pre_middleware"] = aiohttp_pre_middleware
            context["_awssnssqs_message_pre_middleware"] = awssnssqs_message_pre_middleware
            context["_amqp_message_pre_middleware"] = amqp_message_pre_middleware
            context["_schedule_pre_middleware"] = schedule_pre_middleware
            context["_opentelemetry_tracer"] = tracer
            context["_opentelemetry_meter"] = meter

        setattr(service, "_is_instrumented_by_opentelemetry", True)
        setattr(service, "_opentelemetry_tracer", tracer)
        setattr(service, "_opentelemetry_meter", meter)

    @classmethod
    def uninstrument_service(cls, service: tomodachi.Service) -> None:
        if isinstance(service, type):
            raise Exception("uninstrument_service must be called with an instance of tomodachi.Service")

        if cls._instrumented_services is not None:
            try:
                cls._instrumented_services.remove(service)
            except KeyError:
                pass

        aiohttp_pre_middleware = getattr(service, "_aiohttp_pre_middleware", None)
        if aiohttp_pre_middleware:
            for middleware in aiohttp_pre_middleware[:]:
                if isinstance(middleware, OpenTelemetryAioHTTPMiddleware):
                    aiohttp_pre_middleware.remove(middleware)

        awssnssqs_message_pre_middleware = getattr(service, "_awssnssqs_message_pre_middleware", None)
        if awssnssqs_message_pre_middleware:
            for middleware in awssnssqs_message_pre_middleware[:]:
                if isinstance(middleware, OpenTelemetryAWSSQSMiddleware):
                    awssnssqs_message_pre_middleware.remove(middleware)

        amqp_message_pre_middleware = getattr(service, "_amqp_message_pre_middleware", None)
        if amqp_message_pre_middleware:
            for middleware in amqp_message_pre_middleware[:]:
                if isinstance(middleware, OpenTelemetryAMQPMiddleware):
                    amqp_message_pre_middleware.remove(middleware)

        schedule_pre_middleware = getattr(service, "_schedule_pre_middleware", None)
        if schedule_pre_middleware:
            for middleware in schedule_pre_middleware[:]:
                if isinstance(middleware, OpenTelemetryScheduleFunctionMiddleware):
                    schedule_pre_middleware.remove(middleware)

        setattr(service, "_opentelemetry_tracer_provider", None)
        setattr(service, "_opentelemetry_tracer", None)
        setattr(service, "_opentelemetry_meter_provider", None)
        setattr(service, "_opentelemetry_meter", None)
        setattr(service, "_is_instrumented_by_opentelemetry", False)

        context = getattr(service, "context", None)
        if context:
            context["_opentelemetry_tracer"] = None
            context["_opentelemetry_meter"] = None

    @classmethod
    def instrument_logging(
        cls,
        names: Union[Optional[str], List[Optional[str]]],
        logger_provider: Optional[LoggerProvider] = None,
        log_level: int = logging.NOTSET,
    ) -> None:
        logger_provider = cls._logger_provider(logger_provider)

        if not isinstance(names, (list, tuple)):
            names = [names]

        for name in names[:]:
            logger = logging.getLogger(name)
            for handler in logger.handlers:
                if isinstance(handler, OpenTelemetryLoggingHandler):
                    names.remove(name)

        if not names:
            return

        log_level_ = environ.get(OTEL_LOG_LEVEL, log_level) or logging.NOTSET
        if isinstance(log_level_, str) and not log_level_.isdigit():
            log_level = getattr(logging, log_level_.upper(), None) or logging.NOTSET
        elif isinstance(log_level_, str) and log_level_.isdigit():
            log_level = int(log_level_)
        handler = OpenTelemetryLoggingHandler(level=log_level, logger_provider=logger_provider)

        if cls._logging_handlers is None:
            cls._logging_handlers = []
        cls._logging_handlers.append(handler)

        for name in names:
            logging.getLogger(name).addHandler(handler)

    @classmethod
    def uninstrument_logging(cls, names: Union[Optional[str], List[Optional[str]]]) -> None:
        if not isinstance(names, (list, tuple)):
            names = [names]

        for name in names:
            logger = logging.getLogger(name)
            for handler in logger.handlers:
                if not isinstance(handler, OpenTelemetryLoggingHandler):
                    continue
                logger.removeHandler(handler)
                try:
                    handler.acquire()
                    handler.flush()
                    handler.close()
                except (OSError, ValueError):
                    pass
                finally:
                    handler.release()

        if cls._logging_handlers:
            cls._logging_handlers.clear()

    @classmethod
    def instrument_structlog_logger(cls, logger: structlog.BoundLoggerBase) -> None:
        if not isinstance(logger._processors, list):
            raise Exception("logger._processors must be a list")
        if add_trace_structlog_processor not in logger._processors:
            logger._processors.insert(-1, add_trace_structlog_processor)

    @classmethod
    def uninstrument_structlog_logger(cls, logger: structlog.BoundLoggerBase) -> None:
        if not isinstance(logger._processors, list):
            raise Exception("logger._processors must be a list")
        if add_trace_structlog_processor in logger._processors:
            logger._processors.remove(add_trace_structlog_processor)

    def instrumentation_dependencies(self) -> Collection[str]:
        return ("tomodachi",)

    def _instrument_tomodachi(
        self, tracer_provider: TracerProvider, meter_provider: MeterProvider, excluded_urls: Optional[str]
    ) -> None:
        _InstrumentedTomodachiService._opentelemetry_tracer_provider = tracer_provider
        _InstrumentedTomodachiService._opentelemetry_meter_provider = meter_provider
        _InstrumentedTomodachiService._opentelemetry_excluded_urls = excluded_urls

        from tomodachi import TOMODACHI_CLASSES  # isort:skip

        for cls in TOMODACHI_CLASSES:
            if cls is _InstrumentedTomodachiService:
                continue
            for attr in INSTRUMENTATION_CLASS_ATTRIBUTES:
                setattr(cls, attr, getattr(_InstrumentedTomodachiService, attr, None))

        # @todo this wrapping functionality for the publish methods of aws_sns_sqs and amqp could use some refactoring

        # aws_sns_sqs: _publish_message
        if getattr(tomodachi.transport.aws_sns_sqs.AWSSNSSQSTransport._publish_message, "__wrapped__", None):
            setattr(
                tomodachi.transport.aws_sns_sqs.AWSSNSSQSTransport,
                "_publish_message",
                getattr(tomodachi.transport.aws_sns_sqs.AWSSNSSQSTransport._publish_message, "__wrapped__", None),
            )

        aws_sns_sqs_publish_message = tomodachi.transport.aws_sns_sqs.AWSSNSSQSTransport._publish_message

        @functools.wraps(tomodachi.transport.aws_sns_sqs.AWSSNSSQSTransport._publish_message)
        async def _traced_publish_awssnssqs_message(
            cls: tomodachi.transport.aws_sns_sqs.AWSSNSSQSTransport,
            topic_arn: str,
            message: Any,
            message_attributes: Dict,
            context: Dict,
            *args: Any,
            service: Any = None,
            **kwargs: Any,
        ) -> str:
            tracer = (
                cast(Tracer, getattr(service, "_opentelemetry_tracer", None) or context.get("_opentelemetry_tracer"))
                or None
            )
            if not tracer:
                return await aws_sns_sqs_publish_message(
                    topic_arn, message, message_attributes, context, *args, service=service, **kwargs
                )

            topic: str = (
                cls.get_topic_name_without_prefix(cls.decode_topic(cls.get_topic_from_arn(topic_arn)), context)
                if topic_arn
                else ""
            )

            attributes: Dict[str, AttributeValue] = {
                "messaging.system": "aws_sqs",
                "messaging.operation": "publish",
                "messaging.destination.name": topic,
                "messaging.destination.kind": "topic",
            }

            with tracer.start_as_current_span(
                f"{topic} publish",
                kind=SpanKind.PRODUCER,
                attributes=attributes,
            ) as span:
                TraceContextTextMapPropagator().inject(carrier=message_attributes)
                sns_message_id = await aws_sns_sqs_publish_message(
                    topic_arn, message, message_attributes, context, *args, service=service, **kwargs
                )
                span.set_attribute("messaging.message.id", sns_message_id)
                span.set_status(StatusCode.OK)

            return sns_message_id

        setattr(  # noqa: B010
            tomodachi.transport.aws_sns_sqs.AWSSNSSQSTransport,
            "_publish_message",
            _traced_publish_awssnssqs_message.__get__(tomodachi.transport.aws_sns_sqs.AWSSNSSQSTransport),  # type: ignore[attr-defined]
        )

        # aws_sns_sqs: _send_raw_message
        if getattr(tomodachi.transport.aws_sns_sqs.AWSSNSSQSTransport._send_raw_message, "__wrapped__", None):
            setattr(  # noqa: B010
                tomodachi.transport.aws_sns_sqs.AWSSNSSQSTransport,
                "_send_raw_message",
                getattr(tomodachi.transport.aws_sns_sqs.AWSSNSSQSTransport._send_raw_message, "__wrapped__", None),
            )

        aws_sns_sqs_send_raw_message = tomodachi.transport.aws_sns_sqs.AWSSNSSQSTransport._send_raw_message

        @functools.wraps(tomodachi.transport.aws_sns_sqs.AWSSNSSQSTransport._send_raw_message)
        async def _traced_send_raw_awssqs_message(
            cls: tomodachi.transport.aws_sns_sqs.AWSSNSSQSTransport,
            queue_url: str,
            message_body: Any,
            message_attributes: Dict,
            context: Dict,
            *args: Any,
            service: Any = None,
            **kwargs: Any,
        ) -> str:
            tracer = (
                cast(Tracer, getattr(service, "_opentelemetry_tracer", None) or context.get("_opentelemetry_tracer"))
                or None
            )
            if not tracer:
                return await aws_sns_sqs_send_raw_message(
                    queue_url, message_body, message_attributes, context, *args, service=service, **kwargs
                )

            queue_name: str = (
                cls.get_queue_name_without_prefix(cls.get_queue_name_from_queue_url(queue_url), context)
                if queue_url
                else ""
            )

            attributes: Dict[str, AttributeValue] = {
                "messaging.system": "aws_sqs",
                "messaging.operation": "publish",
                "messaging.destination.name": queue_name,
                "messaging.destination.kind": "queue",
            }

            with tracer.start_as_current_span(
                f"{queue_name} publish",
                kind=SpanKind.PRODUCER,
                attributes=attributes,
            ) as span:
                TraceContextTextMapPropagator().inject(carrier=message_attributes)
                sqs_message_id = await aws_sns_sqs_send_raw_message(
                    queue_url, message_body, message_attributes, context, *args, service=service, **kwargs
                )
                span.set_attribute("messaging.message.id", sqs_message_id)
                span.set_status(StatusCode.OK)

            return sqs_message_id

        setattr(  # noqa: B010
            tomodachi.transport.aws_sns_sqs.AWSSNSSQSTransport,
            "_send_raw_message",
            _traced_send_raw_awssqs_message.__get__(tomodachi.transport.aws_sns_sqs.AWSSNSSQSTransport),  # type: ignore[attr-defined]
        )

        # amqp: _publish_message
        if getattr(tomodachi.transport.amqp.AmqpTransport._publish_message, "__wrapped__", None):
            setattr(  # noqa: B010
                tomodachi.transport.amqp.AmqpTransport,
                "_publish_message",
                getattr(tomodachi.transport.amqp.AmqpTransport._publish_message, "__wrapped__", None),
            )

        amqp_publish_message = tomodachi.transport.amqp.AmqpTransport._publish_message

        @functools.wraps(tomodachi.transport.amqp.AmqpTransport._publish_message)
        async def _traced_publish_amqp_message(
            cls: tomodachi.transport.amqp.AmqpTransport,
            routing_key: str,
            exchange_name: str,
            payload: Any,
            properties: Dict,
            routing_key_prefix: Optional[str],
            service: Any,
            context: Dict,
            *args: Any,
            **kwargs: Any,
        ) -> None:
            tracer = (
                cast(Tracer, getattr(service, "_opentelemetry_tracer", None) or context.get("_opentelemetry_tracer"))
                or None
            )
            if not tracer:
                await amqp_publish_message(
                    routing_key,
                    exchange_name,
                    payload,
                    properties,
                    routing_key_prefix,
                    service,
                    context,
                    *args,
                    **kwargs,
                )
                return

            if not exchange_name:
                exchange_name = "amq.topic"

            attributes: Dict[str, AttributeValue] = {
                "messaging.system": "rabbitmq",
                "messaging.operation": "publish",
                "messaging.destination.name": exchange_name,
                "messaging.rabbitmq.destination.routing_key": routing_key,
            }

            if "headers" not in properties:
                properties["headers"] = {}

            with tracer.start_as_current_span(
                f"{routing_key} publish",
                kind=SpanKind.PRODUCER,
                attributes=attributes,
            ) as span:
                TraceContextTextMapPropagator().inject(carrier=properties["headers"])
                await amqp_publish_message(
                    routing_key,
                    exchange_name,
                    payload,
                    properties,
                    routing_key_prefix,
                    service,
                    context,
                    *args,
                    **kwargs,
                )
                span.set_status(StatusCode.OK)

        setattr(  # noqa: B010
            tomodachi.transport.amqp.AmqpTransport,
            "_publish_message",
            _traced_publish_amqp_message.__get__(tomodachi.transport.amqp.AmqpTransport),  # type: ignore[attr-defined]
        )

    def _instrument_logging(self, logger_provider: LoggerProvider) -> None:
        self.instrument_logging(["tomodachi", "exception"], logger_provider=logger_provider)

    def _instrument_structlog_loggers(self) -> None:
        for logger in (
            tomodachi.logging.console_logger,
            tomodachi.logging.no_color_console_logger,
            tomodachi.logging.json_logger,
        ):
            self.instrument_structlog_logger(logger)

    def _instrument(self, **kwargs: Any) -> None:
        tracer_provider: TracerProvider = self._tracer_provider(kwargs.get("tracer_provider"))
        meter_provider: MeterProvider = self._meter_provider(kwargs.get("meter_provider"))
        logger_provider: LoggerProvider = self._logger_provider(kwargs.get("logger_provider"))
        excluded_urls = kwargs.get("excluded_urls")

        self._instrument_tomodachi(tracer_provider, meter_provider, excluded_urls)
        self._instrument_logging(logger_provider)
        self._instrument_structlog_loggers()

    def instrument(self, **kwargs: Any) -> None:
        if not self._is_instrumented_by_opentelemetry or not TomodachiInstrumentor._instrumented_services:
            self._is_instrumented_by_opentelemetry = False
            super().instrument(**kwargs)

    def _uninstrument_tomodachi(self) -> None:
        _InstrumentedTomodachiService._opentelemetry_tracer_provider = None
        _InstrumentedTomodachiService._opentelemetry_meter_provider = None
        _InstrumentedTomodachiService._opentelemetry_excluded_urls = None

        from tomodachi import TOMODACHI_CLASSES  # isort:skip

        for cls in TOMODACHI_CLASSES:
            if cls is _InstrumentedTomodachiService:
                continue
            for attr in INSTRUMENTATION_CLASS_ATTRIBUTES:
                if hasattr(cls, attr):
                    delattr(cls, attr)

        # aws_sns_sqs: _publish_message
        if getattr(tomodachi.transport.aws_sns_sqs.AWSSNSSQSTransport._publish_message, "__wrapped__", None):
            setattr(
                tomodachi.transport.aws_sns_sqs.AWSSNSSQSTransport,
                "_publish_message",
                getattr(tomodachi.transport.aws_sns_sqs.AWSSNSSQSTransport._publish_message, "__wrapped__", None),
            )

        # aws_sns_sqs: _send_raw_message
        if getattr(tomodachi.transport.aws_sns_sqs.AWSSNSSQSTransport._send_raw_message, "__wrapped__", None):
            setattr(
                tomodachi.transport.aws_sns_sqs.AWSSNSSQSTransport,
                "_send_raw_message",
                getattr(tomodachi.transport.aws_sns_sqs.AWSSNSSQSTransport._send_raw_message, "__wrapped__", None),
            )

        # amqp: _publish_message
        if getattr(tomodachi.transport.amqp.AmqpTransport._publish_message, "__wrapped__", None):
            setattr(
                tomodachi.transport.amqp.AmqpTransport,
                "_publish_message",
                getattr(tomodachi.transport.amqp.AmqpTransport._publish_message, "__wrapped__", None),
            )

    def _uninstrument_services(self) -> None:
        if TomodachiInstrumentor._instrumented_services is not None:
            for service in [s for s in TomodachiInstrumentor._instrumented_services]:
                self.uninstrument_service(service)
            TomodachiInstrumentor._instrumented_services.clear()

    def _uninstrument_logging(self) -> None:
        TomodachiInstrumentor.uninstrument_logging(["tomodachi", "exception"])

    def _uninstrument_structlog_loggers(self) -> None:
        for logger in (
            tomodachi.logging.console_logger,
            tomodachi.logging.no_color_console_logger,
            tomodachi.logging.json_logger,
        ):
            TomodachiInstrumentor.uninstrument_structlog_logger(logger)

    def _uninstrument(self, **kwargs: Any) -> None:
        self._uninstrument_tomodachi()
        self._uninstrument_services()
        self._uninstrument_logging()
        self._uninstrument_structlog_loggers()

    def uninstrument(self, **kwargs: Any) -> None:
        if not self._is_instrumented_by_opentelemetry:
            self._is_instrumented_by_opentelemetry = True
        self._uninstrument(**kwargs)

    @staticmethod
    def _tracer_provider(tracer_provider: Optional[TracerProvider] = None) -> TracerProvider:
        if not tracer_provider:
            tracer_provider = _get_tracer_provider() or _create_tracer_provider()
        return tracer_provider

    @staticmethod
    def _meter_provider(
        meter_provider: Optional[MeterProvider] = None, resource: Optional[Resource] = None
    ) -> MeterProvider:
        if not meter_provider:
            meter_provider = _get_meter_provider() or _create_meter_provider()
        _add_meter_provider_views(meter_provider)
        return meter_provider

    @staticmethod
    def _logger_provider(logger_provider: Optional[LoggerProvider] = None) -> LoggerProvider:
        if not logger_provider:
            logger_provider = _get_logger_provider() or _create_logger_provider()
        return logger_provider


class _InstrumentedTomodachiService(tomodachi.Service):
    _tomodachi_class_is_service_class: bool = False

    _is_instrumented_by_opentelemetry: bool = False
    _opentelemetry_tracer_provider: Optional[TracerProvider] = None
    _opentelemetry_meter_provider: Optional[MeterProvider] = None
    _opentelemetry_excluded_urls: Optional[str] = None

    _opentelemetry_tracer: Optional[Tracer] = None
    _opentelemetry_meter: Optional[Meter] = None

    def __post_init_hook(self) -> None:
        TomodachiInstrumentor.instrument_service(
            self,
            tracer_provider=self._opentelemetry_tracer_provider,
            meter_provider=self._opentelemetry_meter_provider,
            excluded_urls=self._opentelemetry_excluded_urls,
        )

    def __post_teardown_hook(self) -> None:
        TomodachiInstrumentor.uninstrument_service(self)



import datetime
from calendar import monthrange
from typing import Any, Dict, List, Optional, Tuple, Union, cast

import pytz

cron_attributes: List[Tuple[str, Tuple[int, int], Dict[str, int]]] = [
    ("minute", (0, 59), {}),
    ("hour", (0, 23), {}),
    ("day", (1, 31), {}),
    (
        "month",
        (1, 12),
        {
            "jan": 1,
            "feb": 2,
            "mar": 3,
            "apr": 4,
            "may": 5,
            "jun": 6,
            "jul": 7,
            "aug": 8,
            "sep": 9,
            "oct": 10,
            "nov": 11,
            "dec": 12,
        },
    ),
    ("isoweekday", (0, 7), {"mon": 1, "tue": 2, "wed": 3, "thu": 4, "fri": 5, "sat": 6, "sun": 0}),
    ("year", (1970, 2099), {}),
]
crontab_aliases: Dict[str, str] = {
    "@yearly": "0 0 1 1 *",
    "@annually": "0 0 1 1 *",
    "@monthly": "0 0 1 * *",
    "@weekly": "0 0 * * 0",
    "@daily": "0 0 * * *",
    "@hourly": "0 * * * *",
    "@minutely": "* * * * *",
}


def get_next_datetime(crontab_notation: str, now_date: datetime.datetime) -> Optional[datetime.datetime]:
    crontab_notation = crontab_aliases.get(crontab_notation, crontab_notation)
    cron_parts = [c for c in crontab_notation.split() if c.strip()]
    cron_parts += ["*" for _ in range(len(cron_attributes) - len(cron_parts))]

    values = []
    last_day = False
    last_weekday = False
    use_weekdays = False
    use_days = False
    for i, attr in enumerate(cron_attributes):
        cron_range: Tuple
        aliases: Dict[str, int]
        _, cron_range, aliases = attr
        available_values: Union[List[int], set] = []
        parts = cron_parts[i].lower().split(",")

        if attr[0] == "isoweekday" and cron_parts[i] != "*":
            use_weekdays = True
        if attr[0] == "day" and cron_parts[i] != "*":
            use_days = True

        for part in parts:
            last = False
            parsed = False
            possible_values: List[int] = [x for x in range(cron_range[0], cron_range[1] + 1)]
            if "-" in part:
                a_value, b_value = part.split("-")
                if "/" in b_value:
                    b_value, _ = b_value.split("/")
                if "l" in a_value[0]:
                    last = True
                    a_value = a_value[1:]
                a = int(aliases.get(a_value, -1))
                b = int(aliases.get(b_value, -1))

                if attr[0] == "isoweekday" and b < a:
                    if b_value == "sun":
                        b = 7
                    else:
                        raise Exception("Invalid cron notation: invalid values for {} ({})".format(attr[0], part))
                if a < 0:
                    try:
                        a = int(a_value)
                    except ValueError as e:
                        raise Exception(
                            "Invalid cron notation: invalid values for {} ({})".format(attr[0], a_value)
                        ) from e
                if b < 0:
                    try:
                        b = int(b_value)
                    except ValueError as e:
                        raise Exception(
                            "Invalid cron notation: invalid values for {} ({})".format(attr[0], b_value)
                        ) from e
                possible_values = [x for x in possible_values if x >= min(a, b) and x <= max(a, b)]
                parsed = True

            if "/" in part:
                a_value, b_value = part.split("/")
                try:
                    a = int(aliases.get(a_value, -1))
                    if a < 0:
                        a = int(a_value)
                    b = int(b_value)
                    possible_values = [x for x in possible_values if x % b == (a % b)]
                except ValueError:
                    try:
                        b = int(b_value)
                    except ValueError as e:
                        raise Exception(
                            "Invalid cron notation: invalid values for {} ({})".format(attr[0], b_value)
                        ) from e
                    if a_value in ["*", "?"]:
                        possible_values = [x for x in possible_values if x % b == 0]
                    else:
                        a_value, _ = part.split("-")
                        a = int(aliases.get(a_value, -1))
                        if a < 0:
                            try:
                                a = int(a_value)
                            except ValueError as e:
                                raise Exception(
                                    "Invalid cron notation: invalid values for {} ({})".format(attr[0], a_value)
                                ) from e
                        possible_values = [x for x in possible_values if x % b == (a % b)]
                parsed = True

            part_value = part
            try:
                if "l" == part_value[0]:
                    last = True
                    part_value = part_value[1:]
                a = int(aliases.get(part_value, -1))
                if a < 0:
                    a = int(part_value)
                possible_values = [x for x in possible_values if x == a]
            except ValueError as e:
                if parsed or part_value in ["*", "?"] or part == "l":
                    pass
                else:
                    raise Exception(
                        "Invalid cron notation: invalid values for {} ({})".format(attr[0], part_value)
                    ) from e

            if last and attr[0] == "day":
                last_day = True
            if last and attr[0] == "isoweekday":
                last_weekday = True

            if attr[0] == "isoweekday":
                possible_values = [x if x != 7 else 0 for x in possible_values]

            if not possible_values:
                raise Exception("Invalid cron notation: invalid values for {}".format(attr[0]))
            if isinstance(available_values, list):
                available_values += possible_values

        available_values = set(available_values)
        values.append(available_values)

    if min(values[2]) >= 28:
        if not any([monthrange(y, m)[1] >= min(values[2]) for y in values[5] for m in values[3]]):
            raise Exception("Invalid cron notation: days out of scope")

    def calculate_date(
        input_date: datetime.datetime, last_day: bool, last_weekday: bool
    ) -> Optional[datetime.datetime]:
        tz: Any = input_date.tzinfo
        naive_date = not bool(input_date.tzinfo)
        if tz is None:
            tz = pytz.UTC

        next_date: Optional[datetime.datetime] = input_date
        while True:
            original_date = next_date
            next_date_weekday = next_date

            for i, attr in enumerate(cron_attributes):
                if attr[0] == "isoweekday":
                    continue
                value = getattr(next_date, attr[0])
                possible_values = [v for v in values[i] if v >= value]
                if not possible_values:
                    if attr[0] == "year":
                        return None
                    next_date = None
                    break
                new_value = min(possible_values)
                try:
                    next_date = datetime.datetime(
                        *[
                            getattr(next_date, dv) if dv != attr[0] else new_value
                            for dv in ["year", "month", "day", "hour", "minute"]
                        ]
                    ).replace(tzinfo=tz)
                except ValueError:
                    next_date = None
                    break

            for i, attr in enumerate(cron_attributes):
                if attr[0] == "isoweekday" or attr[0] == "day":
                    continue
                value = getattr(next_date_weekday, attr[0])
                possible_values = [v for v in values[i] if v >= value]
                if not possible_values:
                    if attr[0] == "year":
                        return None
                    next_date_weekday = None
                    break
                new_value = min(possible_values)
                try:
                    next_date_weekday = datetime.datetime(
                        *[
                            getattr(next_date_weekday, dv) if dv != attr[0] else new_value
                            for dv in ["year", "month", "day", "hour", "minute"]
                        ]
                    ).replace(tzinfo=tz)
                except ValueError:
                    next_date_weekday = None
                    break

            if use_weekdays and next_date_weekday:
                for d in range(
                    next_date_weekday.day, monthrange(next_date_weekday.year, next_date_weekday.month)[1] + 1
                ):
                    next_date_weekday_arguments = (
                        getattr(next_date_weekday, dv) if dv != "day" else d
                        for dv in ["year", "month", "day", "hour", "minute"]
                    )
                    next_date_weekday = datetime.datetime(
                        *cast(Tuple[int, int, int, int, int], next_date_weekday_arguments), tzinfo=tz
                    )
                    if next_date_weekday and (next_date_weekday.isoweekday() % 7) in values[4]:
                        break

            if use_days and not use_weekdays:
                next_date_weekday = None
            if not use_days and use_weekdays:
                next_date = None

            if (
                use_weekdays
                and next_date_weekday
                and (next_date_weekday.isoweekday() % 7) not in values[4]
                and next_date_weekday.isoweekday() not in values[4]
            ):
                next_date_weekday = None
            if next_date and last_day and next_date.day != monthrange(next_date.year, next_date.month)[1]:
                next_date = None
            if use_weekdays and next_date_weekday and last_weekday:
                for i in range(
                    next_date_weekday.day + 1, monthrange(next_date_weekday.year, next_date_weekday.month)[1] + 1
                ):
                    if (
                        datetime.datetime(next_date_weekday.year, next_date_weekday.month, i).isoweekday()
                        == next_date_weekday.isoweekday()
                    ):
                        next_date_weekday = None
                        break

            if use_weekdays and next_date_weekday and (not next_date or next_date_weekday < next_date):
                next_date = next_date_weekday

            if not next_date:
                next_date = original_date
                if next_date:
                    try:
                        next_date = datetime.datetime(next_date.year, next_date.month, next_date.day + 1, tzinfo=tz)
                    except ValueError:
                        try:
                            if next_date:
                                next_date = datetime.datetime(next_date.year, next_date.month + 1, 1, tzinfo=tz)
                        except ValueError:
                            if next_date:
                                next_date = datetime.datetime(next_date.year + 1, 1, 1, tzinfo=tz)
            else:
                break

            if next_date and next_date.year >= 2100:
                return None

        if naive_date and next_date:
            return datetime.datetime(
                next_date.year,
                next_date.month,
                next_date.day,
                next_date.hour,
                next_date.minute,
                next_date.second,
                next_date.microsecond,
            )
        return next_date

    tz: Any = now_date.tzinfo
    calculated_dates = [
        calculate_date(d.replace(tzinfo=tz) if tz else d, last_day, last_weekday)
        for d in [
            (
                datetime.datetime(
                    now_date.year,
                    now_date.month,
                    now_date.day,
                    now_date.hour,
                    now_date.minute,
                    now_date.second,
                    now_date.microsecond,
                )
                if now_date.second == 0
                else None
            ),
            (
                datetime.datetime(now_date.year, now_date.month, now_date.day, now_date.hour, now_date.minute + 1)
                if now_date.minute < 60 - 1
                else None
            ),
            (
                datetime.datetime(now_date.year, now_date.month, now_date.day, now_date.hour + 1)
                if now_date.hour < 24 - 1
                else None
            ),
            (
                datetime.datetime(now_date.year, now_date.month, now_date.day + 1)
                if now_date.day < monthrange(now_date.year, now_date.month)[1] - 1
                else None
            ),
            datetime.datetime(now_date.year, now_date.month + 1, 1) if now_date.month < 12 - 1 else None,
            datetime.datetime(now_date.year + 1, 1, 1),
        ]
        if d
    ]
    if not any(calculated_dates):
        return None
    return min([d for d in calculated_dates if d])



import asyncio
import datetime
import importlib
import os
import platform
import signal
import sys
import traceback
from typing import Any, Callable, Dict, List, Optional, Set, Union, cast

import tomodachi
import tomodachi.container
import tomodachi.importer
import tomodachi.invoker
from tomodachi import logging
from tomodachi.container import ServiceContainer
from tomodachi.helpers.execution_context import clear_execution_context, clear_services, set_execution_context
from tomodachi.helpers.safe_modules import SAFE_MODULES
from tomodachi.importer import ServiceImporter

CancelledError = asyncio.CancelledError
try:
    asyncioexceptions = getattr(asyncio, "exceptions")
    if asyncioexceptions:
        _CancelledError = asyncioexceptions.CancelledError
except (Exception, ModuleNotFoundError, ImportError):
    _CancelledError = asyncio.CancelledError


class ServiceLauncher(object):
    _close_waiter: Optional[asyncio.Future] = None
    _stopped_waiter: Optional[asyncio.Future] = None
    stop_services_pre_hook: Optional[Callable] = None
    stop_services_post_hook: Optional[Callable] = None
    restart_services = False
    services: Set = set()

    @classmethod
    def stop_services(cls) -> None:
        asyncio.ensure_future(cls._stop_services())

    @classmethod
    async def _stop_services(cls) -> None:
        if cls.stop_services_pre_hook:
            await cls.stop_services_pre_hook()
        if cls._close_waiter and not cls._close_waiter.done():
            cls._close_waiter.set_result(None)
            for service in cls.services:
                try:
                    service.stop_service()
                except Exception:
                    pass
            if cls._stopped_waiter:
                cls._stopped_waiter.set_result(None)
        if cls._stopped_waiter:
            await cls._stopped_waiter
        if cls.stop_services_post_hook:
            await cls.stop_services_post_hook()

    @classmethod
    def run_until_complete(
        cls,
        service_files: Union[List, set],
        configuration: Optional[Dict] = None,
        watcher: Any = None,
    ) -> None:
        def sigintHandler(*args: Any) -> None:
            sys.stdout.write("\b\b\r")
            sys.stdout.flush()
            logging.getLogger("tomodachi.signal").warning(
                "interrupt signal <ctrl+c>", signal="SIGINT", process_id=os.getpid()
            )
            cls.restart_services = False

        def sigtermHandler(*args: Any) -> None:
            logging.getLogger("tomodachi.signal").warning(
                "received termination signal", signal="SIGTERM", process_id=os.getpid()
            )
            cls.restart_services = False

        loop: asyncio.AbstractEventLoop
        if sys.version_info.major == 3 and sys.version_info.minor < 10:
            loop = asyncio.get_event_loop()
        else:
            try:
                loop = asyncio.get_running_loop()
            except RuntimeError:
                loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)

        if loop and loop.is_closed():
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)

        for signame in ("SIGINT", "SIGTERM"):
            loop.add_signal_handler(getattr(signal, signame), cls.stop_services)

        signal.siginterrupt(signal.SIGTERM, False)
        signal.siginterrupt(signal.SIGUSR1, False)
        signal.signal(signal.SIGINT, sigintHandler)
        signal.signal(signal.SIGTERM, sigtermHandler)

        watcher_future = None
        if watcher:

            async def _watcher_restart(updated_files: Union[List, set]) -> None:
                cls.restart_services = True

                cwd = os.getcwd()
                for file in service_files:
                    try:
                        ServiceImporter.import_service_file(file, no_exec=True)
                    except (SyntaxError, IndentationError) as e:
                        error_filename = getattr(e, "filename", "")
                        if cwd.rstrip("/") and error_filename.startswith(cwd):
                            error_filename = "./" + error_filename[len(cwd) :].lstrip("/")
                        error_lineno = getattr(e, "lineno", None)
                        error_location = error_filename + (":" + str(error_lineno)) if error_lineno else ""

                        logging.getLogger("tomodachi.watcher").error(
                            "indentation error in file" if type(e) is IndentationError else "syntax error in file",
                            error_location=error_location if error_filename else Ellipsis,
                        )
                        traceback.print_exception(type(e), e, e.__traceback__, limit=0)
                        logging.getLogger("tomodachi.watcher").warning(
                            "restart failed due to error",
                            error_location=error_location if error_filename else Ellipsis,
                        )
                        cls.restart_services = False
                        return
                    except Exception:
                        logging.getLogger("tomodachi.watcher").warning("restart failed due to error")
                        cls.restart_services = False
                        return

                pre_import_current_modules = [m for m in sys.modules.keys()]
                for file in updated_files:
                    if file.lower().endswith(".py"):
                        module_name = file[:-3].replace("/", ".")
                        module_name_full_path = "{}/{}".format(os.path.realpath(cwd), file)[:-3].replace("/", ".")
                        try:
                            for m in pre_import_current_modules:
                                if m == module_name or (len(m) > len(file) and module_name_full_path.endswith(m)):
                                    ServiceImporter.import_module(file, no_exec=True)
                            for m in pre_import_current_modules:
                                if m == module_name or (len(m) > len(file) and module_name_full_path.endswith(m)):
                                    ServiceImporter.import_module(file)
                        except (SyntaxError, IndentationError) as e:
                            error_filename = getattr(e, "filename", "")
                            if cwd.rstrip("/") and error_filename.startswith(cwd):
                                error_filename = "./" + error_filename[len(cwd) :].lstrip("/")
                            error_lineno = getattr(e, "lineno", None)
                            error_location = error_filename + (":" + str(error_lineno)) if error_lineno else ""

                            logging.getLogger("tomodachi.watcher").error(
                                "indentation error in file" if type(e) is IndentationError else "syntax error in file",
                                error_location=error_location if error_filename else Ellipsis,
                            )
                            traceback.print_exception(type(e), e, e.__traceback__, limit=0)
                            logging.getLogger("tomodachi.watcher").warning(
                                "restart failed due to error",
                                error_location=error_location if error_filename else Ellipsis,
                            )
                            cls.restart_services = False
                            return
                        except Exception:
                            logging.getLogger("tomodachi.watcher").warning("restart failed due to error")
                            cls.restart_services = False
                            return

                for file in service_files:
                    try:
                        ServiceImporter.import_service_file(file)
                    except (SyntaxError, IndentationError) as e:
                        error_filename = getattr(e, "filename", "")
                        if cwd.rstrip("/") and error_filename.startswith(cwd):
                            error_filename = "./" + error_filename[len(cwd) :].lstrip("/")
                        error_lineno = getattr(e, "lineno", None)
                        error_location = error_filename + (":" + str(error_lineno)) if error_lineno else ""

                        logging.getLogger("tomodachi.watcher").error(
                            "indentation error in file" if type(e) is IndentationError else "syntax error in file",
                            error_location=error_location if error_filename else Ellipsis,
                        )
                        traceback.print_exception(type(e), e, e.__traceback__, limit=0)
                        logging.getLogger("tomodachi.watcher").warning(
                            "restart failed due to error",
                            error_location=error_location if error_filename else Ellipsis,
                        )
                        cls.restart_services = False
                        return
                    except Exception:
                        logging.getLogger("tomodachi.watcher").warning("restart failed due to error")
                        cls.restart_services = False
                        return

                logging.getLogger("tomodachi.watcher").warning("restarting services")
                cls.stop_services()

            watcher_future = loop.run_until_complete(watcher.watch(loop=loop, callback_func=_watcher_restart))

        cls.restart_services = True
        init_modules = [m for m in sys.modules.keys()]

        restarting = False
        while cls.restart_services:
            if restarting:
                log_level = logging.INFO
                tomodachi.logging.set_default_formatter()
                tomodachi.logging.configure(log_level=log_level)

            init_timestamp_str = (
                datetime.datetime.now(datetime.timezone.utc).isoformat(timespec="microseconds").replace("+00:00", "Z")
            )

            process_id = os.getpid()

            event_loop_alias = ""
            event_loop_version = ""
            try:
                if "uvloop." in str(loop.__class__):
                    event_loop_alias = "uvloop"
                    import uvloop  # noqa  # isort:skip

                    event_loop_version = str(getattr(uvloop, "__version__", ""))
                elif "asyncio." in str(loop.__class__):
                    event_loop_alias = "asyncio"
                else:
                    event_loop_alias = "{}.{}".format(loop.__class__.__module__, loop.__class__.__name__)
            except Exception:
                event_loop_alias = str(loop)

            clear_services()
            clear_execution_context()
            set_execution_context(
                {
                    "tomodachi_version": tomodachi.__version__,
                    "python_version": platform.python_version(),
                    "system_platform": platform.system(),
                    "process_id": process_id,
                    "init_timestamp": init_timestamp_str,
                    "event_loop": event_loop_alias,
                    "cwd": os.getcwd(),
                    "watcher_enabled": True if watcher else False,
                }
            )

            if event_loop_version:
                set_execution_context(
                    {
                        "event_loop_version": event_loop_version,
                    }
                )

            if event_loop_alias == "uvloop" and event_loop_version:
                set_execution_context(
                    {
                        "uvloop_version": event_loop_version,
                    }
                )

            if watcher:
                from tomodachi.helpers.banner import render_banner  # noqa  # isort:skip

                render_banner(service_files=service_files)

            tomodachi.get_contextvar("exit_code").set(-1)
            tomodachi.SERVICE_EXIT_CODE = tomodachi.DEFAULT_SERVICE_EXIT_CODE

            async def _set_waiters() -> None:
                cls._close_waiter = asyncio.Future()
                cls._stopped_waiter = asyncio.Future()

            loop.run_until_complete(_set_waiters())
            cls.restart_services = False

            try:
                cls.services = set(
                    [
                        ServiceContainer(ServiceImporter.import_service_file(file), configuration)
                        for file in service_files
                    ]
                )

                async def _run_until_complete() -> Any:
                    return await asyncio.wait(
                        [asyncio.ensure_future(service.run_until_complete()) for service in cls.services]
                    )

                result = loop.run_until_complete(_run_until_complete())
                exception = [v.exception() for v in [value for value in result if value][0] if v.exception()]
                if exception:
                    raise cast(Exception, exception[0])
                elif restarting and tomodachi.SERVICE_EXIT_CODE and cls._close_waiter and not cls._close_waiter.done():
                    cls.restart_services = True
                    logging.getLogger("tomodachi.watcher").warning("service exited due to errors")
                    logging.getLogger("tomodachi.watcher").warning("trying again in 1.5 seconds")
                    loop.run_until_complete(asyncio.sleep(1.5))

            except tomodachi.importer.ServicePackageError:
                pass
            except (SyntaxError, IndentationError) as e:
                cwd = os.getcwd()
                error_filename = getattr(e, "filename", "")
                if error_filename.startswith(cwd):
                    error_filename = error_filename[len(cwd) :].lstrip("/")
                error_lineno = getattr(e, "lineno", None)
                error_location = error_filename + (":" + str(error_lineno)) if error_lineno else ""

                logging.getLogger("exception").error(
                    "indentation error in file" if type(e) is IndentationError else "syntax error in file",
                    error_location=error_location if error_filename else Ellipsis,
                )
                traceback.print_exception(type(e), e, e.__traceback__, limit=0)
                if not cls.restart_services:
                    tomodachi.SERVICE_EXIT_CODE = 1
            except Exception as e:
                logging.getLogger("exception").exception("uncaught exception: {}".format(str(e)))

                if isinstance(e, ModuleNotFoundError):  # pragma: no cover
                    missing_module_name = str(getattr(e, "name", None) or "")
                    if missing_module_name:
                        color = ""
                        color_reset = ""
                        try:
                            import colorama  # noqa  # isort:skip

                            color = colorama.Fore.WHITE + colorama.Back.RED
                            color_reset = colorama.Style.RESET_ALL
                        except Exception:
                            pass

                        print("")
                        print(
                            "{}[fatal error] The '{}' package is missing or cannot be imported.{}".format(
                                color, missing_module_name, color_reset
                            )
                        )
                        print("")

                if restarting:
                    logging.getLogger("tomodachi.watcher").warning("cannot restart due to errors")
                    logging.getLogger("tomodachi.watcher").warning("trying again in 1.5 seconds")
                    loop.run_until_complete(asyncio.sleep(1.5))
                    if cls._close_waiter and not cls._close_waiter.done():
                        cls.restart_services = True
                    else:
                        for signame in ("SIGINT", "SIGTERM"):
                            loop.remove_signal_handler(getattr(signal, signame))
                else:
                    for signame in ("SIGINT", "SIGTERM"):
                        loop.remove_signal_handler(getattr(signal, signame))

                if not cls.restart_services:
                    tomodachi.SERVICE_EXIT_CODE = 1

            if cls.restart_services:
                # log handler cleanup
                logging.remove_handlers()
                tomodachi.logging.reset_context()

            current_modules = [m for m in sys.modules.keys()]
            for m in current_modules:
                if m not in init_modules and m not in SAFE_MODULES:
                    del sys.modules[m]

            importlib.reload(tomodachi.container)
            importlib.reload(tomodachi.invoker)
            importlib.reload(tomodachi.invoker.base)
            importlib.reload(tomodachi.importer)

            restarting = True

        if watcher:
            if watcher_future and not watcher_future.done():
                try:
                    watcher_future.set_result(None)
                except RuntimeError:  # pragma: no cover
                    watcher_future.cancel()
                if not watcher_future.done():  # pragma: no cover
                    try:
                        loop.run_until_complete(watcher_future)
                    except (Exception, CancelledError, _CancelledError):
                        pass



import base64
import json
import time
import uuid
import zlib
from typing import Any, Dict, Tuple, Union

PROTOCOL_VERSION = "tomodachi-json-base--1.0.0"


class JsonBase(object):
    @classmethod
    async def build_message(cls, service: Any, topic: str, data: Any, **kwargs: Any) -> str:
        data_encoding = "raw"
        if len(json.dumps(data)) >= 60000:
            data = base64.b64encode(zlib.compress(json.dumps(data).encode("utf-8"))).decode("utf-8")
            data_encoding = "base64_gzip_json"

        message = {
            "service": {"name": getattr(service, "name", None), "uuid": getattr(service, "uuid", None)},
            "metadata": {
                "message_uuid": "{}.{}".format(getattr(service, "uuid", ""), str(uuid.uuid4())),
                "protocol_version": PROTOCOL_VERSION,
                "compatible_protocol_versions": ["json_base-wip"],  # deprecated
                "timestamp": time.time(),
                "topic": topic,
                "data_encoding": data_encoding,
            },
            "data": data,
        }
        return json.dumps(message)

    @classmethod
    async def parse_message(cls, payload: str, **kwargs: Any) -> Union[Dict, Tuple]:
        message = json.loads(payload)

        message_uuid = message.get("metadata", {}).get("message_uuid")
        timestamp = message.get("metadata", {}).get("timestamp")

        data = None
        if message.get("metadata", {}).get("data_encoding") == "raw":
            data = message.get("data")
        elif message.get("metadata", {}).get("data_encoding") == "base64_gzip_json":
            data = json.loads(zlib.decompress(base64.b64decode(message.get("data").encode("utf-8"))).decode("utf-8"))

        return (
            {
                "service": {
                    "name": message.get("service", {}).get("name"),
                    "uuid": message.get("service", {}).get("uuid"),
                },
                "metadata": {
                    "message_uuid": message.get("metadata", {}).get("message_uuid"),
                    "protocol_version": message.get("metadata", {}).get("protocol_version"),
                    "timestamp": message.get("metadata", {}).get("timestamp"),
                    "topic": message.get("metadata", {}).get("topic"),
                    "data_encoding": message.get("metadata", {}).get("data_encoding"),
                },
                "data": data,
            },
            message_uuid,
            timestamp,
        )


__all__ = [
    "PROTOCOL_VERSION",
    "JsonBase",
]



from typing import Any, Dict, Optional, cast

_services: Dict[str, Any] = {}
_current_service: Dict[int, Any] = {}
_context: Dict[str, Any] = {}


def set_service(name: str, instance: Any) -> None:
    _services[name] = instance
    _current_service[0] = instance


def unset_service(name: str) -> None:
    del _services[name]


def clear_services() -> None:
    for name in list(_services.keys()):
        del _services[name]
    if _current_service:
        del _current_service[0]


def get_service(name: Optional[str] = None) -> Any:
    if name is None:
        if _current_service and len(_current_service):
            return _current_service[0]

        for k, v in _services.items():
            name = k
            break

    return _services.get(name) if name else None


def get_services() -> Any:
    return list(_services.values())


def get_instance(name: Optional[str] = None) -> Any:
    # alias for tomodachi.get_service()
    return get_service(name)


def set_execution_context(values: Dict[str, Any]) -> None:
    for key, value in values.items():
        _context[key] = value


def get_execution_context() -> Dict[str, Any]:
    return _context


def clear_execution_context() -> None:
    for key in list(_context.keys()):
        del _context[key]


def increase_execution_context_value(key: str, value: int = 1) -> int:
    if key not in _context:
        _context[key] = 0
    if not isinstance(_context[key], int):
        raise Exception("Cannot increase non-integer context value")
    _context[key] += value
    return cast(int, _context[key])


def decrease_execution_context_value(key: str, value: int = 1) -> int:
    return increase_execution_context_value(key, -value)



from tomodachi.transport import amqp as amqp
from tomodachi.transport import aws_sns_sqs as aws_sns_sqs
from tomodachi.transport import awssnssqs as awssnssqs
from tomodachi.transport import http as http
from tomodachi.transport import schedule as schedule

__all__ = ["amqp", "aws_sns_sqs", "awssnssqs", "http", "schedule"]



from tomodachi.discovery import aws_sns_registration as aws_sns_registration
from tomodachi.discovery import dummy_registry as dummy_registry
from tomodachi.discovery.aws_sns_registration import AWSSNSRegistration as AWSSNSRegistration
from tomodachi.discovery.aws_sns_registration import AWSSNSRegistration as aws_sns
from tomodachi.discovery.aws_sns_registration import AWSSNSRegistration as awssns
from tomodachi.discovery.dummy_registry import DummyRegistry as DummyRegistry
from tomodachi.discovery.dummy_registry import DummyRegistry as dummy
from tomodachi.discovery.dummy_registry import DummyRegistry as example

__all__ = [
    "AWSSNSRegistration",
    "awssns",
    "aws_sns",
    "DummyRegistry",
    "dummy",
    "example",
    "aws_sns_registration",
    "dummy_registry",
]



#!/usr/bin/env python
import asyncio
import getopt
import logging
import os
import sys
from typing import Dict, List, Literal, Optional, Tuple, Union, cast

import tomodachi
from tomodachi.config import parse_config_files
from tomodachi.helpers.build_time import get_time_since_build
from tomodachi.launcher import ServiceLauncher


class CLI:
    def help_command_usage(self) -> str:
        from tomodachi.helpers.colors import COLOR, COLOR_RESET, COLOR_STYLE

        LABEL = f"{COLOR_RESET}{COLOR.YELLOW}{COLOR_STYLE.BRIGHT}"
        SHELL = f"{COLOR_RESET}{COLOR.WHITE}{COLOR_STYLE.DIM}"
        MAIN_USAGE = f"{COLOR_RESET}{COLOR.WHITE}{COLOR_STYLE.BRIGHT}"
        AVAILABLE_COMMAND = f"{COLOR_RESET}{COLOR.BLUE}"
        OPTION = f"{COLOR_RESET}{COLOR.GREEN}"
        DEFAULT = f"{COLOR_RESET}{COLOR.WHITE}{COLOR_STYLE.DIM}"
        BOTTOM_TEXT = f"{COLOR_RESET}{COLOR.LIGHTBLACK_EX}"
        BOTTOM_LABEL = f"{COLOR_RESET}{COLOR.WHITE}{COLOR_STYLE.DIM}"

        time_since_tomodachi_build = get_time_since_build() or "local development version"
        return (
            f"{LABEL}usage:{COLOR_RESET}\n"
            f"  {SHELL}${COLOR_RESET} {MAIN_USAGE}tomodachi run [options] <service.py ...>{COLOR_RESET}\n"
            "\n"
            f"{LABEL}description:{COLOR_RESET}\n"
            f"  starts the tomodachi service(s) defined in the files provided as arguments.\n"
            "\n"
            f"{LABEL}options:{COLOR_RESET}\n"
            f"  {OPTION}--loop [auto|asyncio|uvloop]{COLOR_RESET}\n"
            f"      use the specified event loop implementation. {DEFAULT}(default: auto){COLOR_RESET}\n"
            f"  {OPTION}--production{COLOR_RESET}\n"
            "      disables service restart on file changes and hides the info banner.\n"
            f"  {OPTION}--log-level [debug|info|warning|error|critical]{COLOR_RESET}\n"
            f"      specify the minimum log level. {DEFAULT}(default: info){COLOR_RESET}\n"
            f"  {OPTION}--logger [console|json|python|disabled]{COLOR_RESET}\n"
            f"      specify a log formatter for tomodachi.logging. {DEFAULT}(default: console){COLOR_RESET}\n"
            f"  {OPTION}--custom-logger <module.attribute|module>{COLOR_RESET}\n"
            "      use a custom logger object or custom log module (as import path).\n"
            f"  {OPTION}--opentelemetry-instrument{COLOR_RESET}\n"
            "      auto-instruments the application with opentelemetry instrumentors.\n"
            "\n"
            f"{LABEL}usage examples:{COLOR_RESET}\n"
            f"  {SHELL}${COLOR_RESET} {AVAILABLE_COMMAND}tomodachi run --production --logger json --loop uvloop service/app.py{COLOR_RESET}\n"
            f"  {SHELL}${COLOR_RESET} {AVAILABLE_COMMAND}tomodachi run --log-level warning --custom-logger foobar.logger service.py{COLOR_RESET}\n"
            "\n"
            f"{BOTTOM_LABEL}ver{COLOR_RESET} {BOTTOM_TEXT}{tomodachi.__version__} ({time_since_tomodachi_build}){COLOR_RESET}\n"
            f"{BOTTOM_LABEL}git{COLOR_RESET} {BOTTOM_TEXT}https://github.com/kalaspuff/tomodachi{COLOR_RESET}\n"
        )

    def help_command(self) -> None:
        print(self.help_command_usage())
        sys.exit(0)

    def version_command(self) -> None:
        print("tomodachi {}".format(tomodachi.__version__))
        sys.exit(0)

    def dependency_versions_command(self) -> None:
        CLI.test_dependencies(fail_on_errors=False, output_versions=True, output_errors=True)
        sys.exit(0)

    @classmethod
    def test_dependencies(
        cls, fail_on_errors: bool = True, output_versions: bool = False, output_errors: bool = True
    ) -> Dict[str, Optional[str]]:
        errors = False

        aioamqp_version = ""
        aiobotocore_version = ""
        aiohttp_version = ""
        botocore_version = ""
        protobuf_version = ""
        uvloop_version = ""
        opentelemetry_api_version = ""
        opentelemetry_sdk_version = ""
        opentelemetry_instrumentation_version = ""

        try:
            import aioamqp  # noqa  # isort:skip

            aioamqp_version = aioamqp.__version__
            if output_versions:
                print("aioamqp/{}".format(aioamqp_version))
        except ModuleNotFoundError as e:  # pragma: no cover
            errors = True
            if output_errors:
                print("Dependency failure: 'aioamqp' failed to load (error: \"{}\")".format(str(e)))
        except Exception as e:  # pragma: no cover
            errors = True
            if output_errors:
                print("Dependency failure: 'aioamqp' failed to load (error: \"{}\")".format(str(e)))
                logging.exception("")
                print("")

        try:
            import aiobotocore  # noqa  # isort:skip

            aiobotocore_version = aiobotocore.__version__
            if output_versions:
                print("aiobotocore/{}".format(aiobotocore_version))
        except ModuleNotFoundError as e:  # pragma: no cover
            errors = True
            if output_errors:
                print("Dependency failure: 'aiobotocore' failed to load (error: \"{}\")".format(str(e)))
        except Exception as e:  # pragma: no cover
            errors = True
            if output_errors:
                print("Dependency failure: 'aiobotocore' failed to load (error: \"{}\")".format(str(e)))
                logging.exception("")
                print("")

        try:
            import aiohttp  # noqa  # isort:skip

            aiohttp_version = aiohttp.__version__
            if output_versions:
                print("aiohttp/{}".format(aiohttp_version))
        except ModuleNotFoundError as e:  # pragma: no cover
            errors = True
            if output_errors:
                print("Dependency failure: 'aiohttp' failed to load (error: \"{}\")".format(str(e)))
        except Exception as e:  # pragma: no cover
            errors = True
            if output_errors:
                print("Dependency failure: 'aiohttp' failed to load (error: \"{}\")".format(str(e)))
                logging.exception("")
                print("")

        try:
            import botocore  # noqa  # isort:skip

            botocore_version = botocore.__version__
            if output_versions:
                print("botocore/{}".format(botocore_version))
        except ModuleNotFoundError as e:  # pragma: no cover
            errors = True
            if output_errors:
                print("Dependency failure: 'botocore' failed to load (error: \"{}\")".format(str(e)))
        except Exception as e:  # pragma: no cover
            errors = True
            if output_errors:
                print("Dependency failure: 'botocore' failed to load (error: \"{}\")".format(str(e)))
                logging.exception("")
                print("")

        try:
            # Optional
            import google.protobuf  # noqa  # isort:skip

            protobuf_version_ = google.protobuf.__version__
            if isinstance(protobuf_version_, bytes):  # type: ignore
                protobuf_version = cast(bytes, protobuf_version_).decode()  # type: ignore
            else:
                protobuf_version = str(protobuf_version_)
            if output_versions:
                print("protobuf/{}".format(protobuf_version))
        except ModuleNotFoundError:  # pragma: no cover
            pass
        except Exception:  # pragma: no cover
            pass

        try:
            # Optional
            import uvloop  # noqa  # isort:skip

            uvloop_version = str(getattr(uvloop, "__version__", ""))
            if output_versions:
                print("uvloop/{}".format(uvloop_version))
        except ModuleNotFoundError:  # pragma: no cover
            pass
        except Exception:  # pragma: no cover
            pass

        try:
            # Optional
            import opentelemetry.instrumentation.version  # noqa  # isort:skip
            import opentelemetry.sdk.version  # noqa  # isort:skip
            import opentelemetry.version  # noqa  # isort:skip

            opentelemetry_api_version = opentelemetry.version.__version__
            opentelemetry_sdk_version = opentelemetry.sdk.version.__version__
            opentelemetry_instrumentation_version = opentelemetry.instrumentation.version.__version__
            if output_versions:
                print("opentelemetry-api/{}".format(opentelemetry_api_version))
                print("opentelemetry-sdk/{}".format(opentelemetry_sdk_version))
                print("opentelemetry-instrumentation/{}".format(opentelemetry_instrumentation_version))
        except ModuleNotFoundError:  # pragma: no cover
            pass
        except Exception:  # pragma: no cover
            pass

        if not errors:
            try:
                import tomodachi.helpers.logging  # noqa  # isort:skip
                import tomodachi.invoker  # noqa  # isort:skip
                import tomodachi.transport.amqp  # noqa  # isort:skip
                import tomodachi.transport.aws_sns_sqs  # noqa  # isort:skip
                import tomodachi.transport.http  # noqa  # isort:skip
                import tomodachi.transport.schedule  # noqa  # isort:skip
            except Exception as e:  # pragma: no cover
                errors = True
                if output_errors:
                    print('Dependency failure: tomodachi essentials failed to load (error: "{}")'.format(str(e)))
                    logging.exception("")
                    print("")

        if errors:
            if fail_on_errors:
                logging.getLogger("exception").warning("Unable to initialize dependencies")
                logging.getLogger("exception").warning("Error: See above exceptions and traceback")
                sys.exit(1)
            elif output_errors:
                print("There were errors - see above for exceptions and traceback")

        return {
            "aioamqp": aioamqp_version or None,
            "aiobotocore": aiobotocore_version or None,
            "aiohttp": aiohttp_version or None,
            "botocore": botocore_version or None,
            "protobuf": protobuf_version or None,
            "uvloop": uvloop_version or None,
            "opentelemetry-api": opentelemetry_api_version or None,
            "opentelemetry-sdk": opentelemetry_sdk_version or None,
            "opentelemetry-instrumentation": opentelemetry_instrumentation_version or None,
        }

    def run_command(self, args: List[str]) -> None:
        if len(args) == 0:
            print(self.help_command_usage())
        else:
            configuration = None
            log_level = logging.INFO

            tomodachi.get_contextvar("run.args").set(args[:])

            # --loop (env: TOMODACHI_LOOP)
            env_loop = str(os.getenv("TOMODACHI_LOOP", "")).lower() or None

            if env_loop or "--loop" in args:
                if "--loop" in args:
                    index = args.index("--loop")
                    args.pop(index)
                    value = args.pop(index).lower()

                    if env_loop and env_loop != value:
                        print("Invalid value for --loop option: '{}' differs from env TOMODACHI_LOOP".format(value))
                        sys.exit(2)
                elif env_loop:
                    value = env_loop
                else:
                    value = "auto"

                if value in ("auto", "default"):
                    value = "auto"
                elif value in ("asyncio", "aio", "async"):
                    asyncio.set_event_loop_policy(asyncio.DefaultEventLoopPolicy())
                    value = "asyncio"
                elif value in ("uvloop", "libuv", "uv"):
                    try:
                        import uvloop  # noqa  # isort:skip
                    except Exception:  # pragma: no cover
                        print("The 'uvloop' package needs to be installed to use uvloop event loop")
                        sys.exit(2)
                    asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())
                    value = "uvloop"
                else:
                    print("Invalid value for --loop option: '{}' not recognized".format(value))
                    sys.exit(2)

                tomodachi.get_contextvar("loop.setting").set(value)
            else:
                tomodachi.get_contextvar("loop.setting").set("auto")

            # --config
            if "-c" in args or "--config" in args:
                import warnings  # isort:skip

                warnings.warn(
                    "Using the -c (--config) CLI argument is deprecated. Set and parse service config with environment variables instead.",
                    DeprecationWarning,
                )

                index = args.index("-c") if "-c" in args else args.index("--config")
                args.pop(index)

                config_files: List[str] = []
                while len(args) > index and args[index][0] != "-":
                    value = args.pop(index)
                    if value not in config_files:
                        config_files.append(value)

                if not len(config_files):
                    print("Missing config file on command line")
                    sys.exit(2)

                try:
                    configuration = parse_config_files(config_files)
                except FileNotFoundError as e:
                    print("Invalid config file: {}".format(str(e)))
                    sys.exit(2)
                except ValueError as e:
                    print("Invalid config file, invalid JSON format: {}".format(str(e)))
                    sys.exit(2)

            # --production (env: TOMODACHI_PRODUCTION)
            env_production = str(os.getenv("TOMODACHI_PRODUCTION", "")).lower() or None
            if env_production and env_production in ("0", "no", "none", "false"):
                env_production = None

            if env_production or "--production" in args:
                if "--production" in args:
                    index = args.index("--production")
                    args.pop(index)
                watcher = None
            else:
                cwd = os.path.realpath(os.getcwd())
                root_directories = [cwd]
                for arg in set(args):
                    if not arg.startswith("/") and not arg.startswith("~"):
                        root_directories.append(os.path.realpath(os.path.dirname(os.path.join(cwd, arg))))
                    else:
                        root_directories.append(os.path.realpath(os.path.dirname(arg)))
                for p in str(os.getenv("PYTHONPATH", "")).split(os.pathsep):
                    if not p:
                        continue
                    if not p.startswith("/") and not p.startswith("~"):
                        root_directories.append(os.path.realpath(os.path.join(cwd, p)))
                    else:
                        root_directories.append(os.path.realpath(p))

                from tomodachi.watcher import Watcher  # noqa  #  isort:skip

                watcher = Watcher(root=root_directories, configuration=configuration)

            # --log-level (env: TOMODACHI_LOG_LEVEL)
            env_log_level: Optional[Union[str, int]] = str(os.getenv("TOMODACHI_LOG_LEVEL", "")).lower() or None
            if not env_log_level:
                env_log_level = None

            if env_log_level is not None:
                log_level_ = getattr(logging, str(env_log_level).upper(), None) or logging.NOTSET
                if type(log_level_) is not int or log_level_ == logging.NOTSET:
                    print(
                        "Invalid TOMODACHI_LOG_LEVEL environment value (expected 'debug', 'info', 'warning', 'error' or 'critical')"
                    )
                    sys.exit(2)
                env_log_level = log_level_

            if "-l" in args or "--log" in args or "--log-level" in args:
                index = (
                    args.index("-l")
                    if "-l" in args
                    else (args.index("--log") if "--log" in args else args.index("--log-level"))
                )
                args.pop(index)
                if len(args) > index:
                    arg_value = args.pop(index)
                    log_level_ = getattr(logging, arg_value.upper(), None) or logging.NOTSET
                    if type(log_level_) is not int or log_level_ == logging.NOTSET:
                        print(
                            "Invalid log level: '{}' (expected 'debug', 'info', 'warning', 'error' or 'critical')".format(
                                arg_value
                            )
                        )
                        sys.exit(2)

                    if env_log_level is not None and env_log_level != log_level_:
                        print(
                            "Invalid value for --log-level option: '{}' differs from env TOMODACHI_LOG_LEVEL".format(
                                value
                            )
                        )
                        sys.exit(2)

                    log_level = log_level_
            elif env_log_level and type(env_log_level) is int:
                log_level = env_log_level

            # --logger (env: TOMODACHI_LOGGER)
            env_logger = cast(
                Optional[Literal["json", "console", "custom", "python", "disabled"]],
                str(os.getenv("TOMODACHI_LOGGER", "")).lower() or None,
            )
            if not env_logger:
                env_logger = None

            if env_logger or "--logger" in args:
                if "--logger" in args:
                    index = args.index("--logger")
                    args.pop(index)
                    try:
                        value = args.pop(index).lower()
                    except IndexError:
                        print("Missing value for --logger option")
                        sys.exit(2)

                    if env_logger and env_logger != value:
                        print("Invalid value for --logger option: '{}' differs from env TOMODACHI_LOGGER".format(value))
                        sys.exit(2)
                else:
                    value = env_logger or ""

                if value not in ("json", "console", "custom", "python", "disabled"):
                    print(
                        "Invalid value for --logger option: '{}' (expected 'json', 'console', 'python' or 'disabled')".format(
                            value
                        )
                    )
                    sys.exit(2)

                env_logger = cast(Literal["json", "console", "custom", "python", "disabled"], value)

            if not env_logger:
                env_logger = None

            # --custom-logger (env: TOMODACHI_CUSTOM_LOGGER)
            env_custom_logger = str(os.getenv("TOMODACHI_CUSTOM_LOGGER", "")).lower() or None
            if not env_custom_logger:
                env_custom_logger = None

            if env_custom_logger or "--custom-logger" in args:
                if "--custom-logger" in args:
                    index = args.index("--custom-logger")
                    args.pop(index)
                    try:
                        value = args.pop(index)
                    except IndexError:
                        print("Missing value for --custom-logger option")
                        sys.exit(2)

                    if env_custom_logger and env_custom_logger != value:
                        print(
                            "Invalid value for --custom-logger option: '{}' differs from env TOMODACHI_CUSTOM_LOGGER".format(
                                value
                            )
                        )
                        sys.exit(2)
                else:
                    value = env_custom_logger or ""

                if env_logger and env_logger != "custom":
                    print("Invalid combination of --custom-logger and --logger options")
                    sys.exit(2)

                env_custom_logger = value
                env_logger = "custom"

            if env_logger and env_logger == "custom" and not env_custom_logger:
                print("Invalid combination of --custom-logger and --logger options")
                sys.exit(2)

            # --opentelemetry-instrument (env: TOMODACHI_OPENTELEMETRY_INSTRUMENT)
            env_opentelemetry_instrument = str(os.getenv("TOMODACHI_OPENTELEMETRY_INSTRUMENT", "")).lower() or None
            if env_opentelemetry_instrument and env_opentelemetry_instrument in ("0", "no", "none", "false"):
                env_opentelemetry_instrument = None

            if env_opentelemetry_instrument or "--opentelemetry-instrument" in args:
                if "--opentelemetry-instrument" in args:
                    index = args.index("--opentelemetry-instrument")
                    args.pop(index)

                try:
                    import opentelemetry.instrumentation.auto_instrumentation._load  # noqa  # isort:skip
                    import opentelemetry.util.http  # noqa  # isort:skip
                    import opentelemetry.sdk.version  # noqa  # isort:skip
                    import opentelemetry.version  # noqa  # isort:skip
                except Exception:  # pragma: no cover
                    print("The 'opentelemetry' extras needs to be installed to use OpenTelemetry instrumentation")
                    sys.exit(2)

                auto_instrument_opentelemetry = True
            else:
                auto_instrument_opentelemetry = False

            if not args:
                from tomodachi.helpers.colors import COLOR, COLOR_RESET, COLOR_STYLE

                print(f"{COLOR.RED}error:{COLOR_RESET} no service file has been provided as argument.")

                print("")
                print(f"{COLOR.WHITE}{COLOR_STYLE.DIM}---{COLOR_RESET}")
                print("")

                print("use the '--help' option for cli usage help.")
                print(f"{COLOR.WHITE}{COLOR_STYLE.DIM}${COLOR_RESET} {COLOR.BLUE}tomodachi --help{COLOR_RESET}")
                sys.exit(2)

            if auto_instrument_opentelemetry:
                from tomodachi.opentelemetry.auto_instrumentation import initialize as initialize_opentelemetry

                try:
                    initialize_opentelemetry()
                except Exception:
                    logging.getLogger("tomodachi.opentelemetry").exception("Failed to auto initialize opentelemetry")

            tomodachi.logging.set_default_formatter(env_logger)
            tomodachi.logging.set_custom_logger_factory(env_custom_logger)
            tomodachi.logging.configure(log_level=log_level)

            ServiceLauncher.run_until_complete(set(args), configuration, watcher)

        # Cleanup log handlers
        tomodachi.logging.remove_handlers()

        sys.exit(tomodachi.SERVICE_EXIT_CODE)

    def main(self, argv: List[str]) -> None:
        opts: List[Tuple[str, str]] = []
        args: List[str] = []

        try:
            opts, args = getopt.getopt(
                argv,
                "hlvV ",
                [
                    "help",
                    "log",
                    "version",
                    "version",
                    "dependency-versions",
                    "dependencies",
                    "deps",
                    "logger",
                    "log-level",
                    "custom-logger",
                    "production",
                    "loop",
                ],
            )
        except getopt.GetoptError as e:
            from tomodachi.helpers.colors import COLOR, COLOR_RESET, COLOR_STYLE

            print(f"{COLOR.RED}error:{COLOR_RESET} invalid command or combination of command options.")
            print(f"{COLOR.RED}error:{COLOR_RESET} {str(e)}.")

            print("")
            print(f"{COLOR.WHITE}{COLOR_STYLE.DIM}---{COLOR_RESET}")
            print("")

            print("use the '--help' option for cli usage help.")
            print(f"{COLOR.WHITE}{COLOR_STYLE.DIM}${COLOR_RESET} {COLOR.BLUE}tomodachi --help{COLOR_RESET}")
            sys.exit(2)
        for opt, _ in opts:
            if opt in ("-h", "--help"):
                self.help_command()
            if opt in ("-v", "-V", "--version"):
                self.version_command()
            if opt in ("--dependency-versions", "--dependencies", "--deps"):
                self.dependency_versions_command()

            if opt in ("-l", "--log-level", "--log", "--logger", "--custom-logger", "--production", "--loop"):
                from tomodachi.helpers.colors import COLOR, COLOR_RESET, COLOR_STYLE

                print(f"{COLOR.RED}error:{COLOR_RESET} invalid command or combination of command options.")
                print(f"{COLOR.RED}error:{COLOR_RESET} the command 'run' must be specified before any options.")

                print("")
                print(f"{COLOR.WHITE}{COLOR_STYLE.DIM}---{COLOR_RESET}")
                print("")

                if any([a.endswith(".py") for a in argv]):
                    new_args = ["run"] + [a for a in argv if a not in ("run", "start", "go")]
                    print("maybe you intended to run something like this?")
                    print(
                        f"{COLOR.WHITE}{COLOR_STYLE.DIM}${COLOR_RESET} {COLOR.BLUE}tomodachi {' '.join(new_args)}{COLOR_RESET}"
                    )
                    print("")

                print("use the '--help' option for cli usage help.")
                print(f"{COLOR.WHITE}{COLOR_STYLE.DIM}${COLOR_RESET} {COLOR.BLUE}tomodachi --help{COLOR_RESET}")
                sys.exit(2)

        if len(args):
            if args[0] in ("run", "start", "go"):
                self.run_command(args[1:])

        if args or opts:
            from tomodachi.helpers.colors import COLOR, COLOR_RESET, COLOR_STYLE

            print(f"{COLOR.RED}error:{COLOR_RESET} invalid command or combination of command options.")
            print(
                f"{COLOR.RED}error:{COLOR_RESET} the command 'run' must be specified before any service files or options."
            )

            print("")
            print(f"{COLOR.WHITE}{COLOR_STYLE.DIM}---{COLOR_RESET}")
            print("")

            if any([a.endswith(".py") for a in argv]):
                new_args = ["run"] + [a for a in argv]
                print("maybe you intended to run something like this?")
                print(
                    f"{COLOR.WHITE}{COLOR_STYLE.DIM}${COLOR_RESET} {COLOR.BLUE}tomodachi {' '.join(new_args)}{COLOR_RESET}"
                )
                print("")

            print("use the '--help' option for cli usage help.")
            print(f"{COLOR.WHITE}{COLOR_STYLE.DIM}${COLOR_RESET} {COLOR.BLUE}tomodachi --help{COLOR_RESET}")
            sys.exit(2)

        self.help_command()


def cli_entrypoint(argv: Optional[List[str]] = None) -> None:
    if argv is None:
        argv = sys.argv
        if argv[0].endswith("pytest"):
            argv = ["tomodachi"]

    CLI().main(argv[1:])
